/* Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 *  For details, see the files COPYING and LICENSE in the base directory
 *  of the package.
 */
#ifndef SCADOPT_HPP_
#define SCADOPT_HPP_

/*******************************************************************************
 ******************************************************************************/
/**
 *  @file SCADOPT.hpp
 *
 *  @brief Main include file for applications using SCADOPT.
 *
 * This file has to be included if a new PDE constrained optimization problem
 * should be created.
 */

/**
 *  \mainpage SCADOPT Documentation
 *
 * SCADOPT is a highly-scalable open-source HPC framework written in C++ for
 * solving PDE constrained optimization problems. The framework is designed
 * to handle explicit time-stepping methods over structured grids for the
 * approximation of time-dependent PDEs.
 *
 * \n It provides an interface to gradient-based optimization algorithms like
 * L-BFGS-B. Derivatives needed for the optimization are computed via
 * algorithmic differentiation (AD) since this approach provides exact
 * derivatives values with a reasonable computational overhead.
 * Hence, the integration of the AD-tool ADOL-C has been a major design
 * principle for the data management of SCADOPT.
 *
 * The framework is freely available under the new BSD license. For details,
 * please refer to the files COPYING and LICENSE in the base directory
 * of the package.
 *
 * \section Implementation How to Implement a New Problem?
 * The usual way to implement a new discretised problem is as follows:
 * - Create the skeleton of the discretized problem.
 * The skeleton of the discretised problem can be setup by creating a new
 * class which is derived from the class \c SCADOPT::Problem.
 * - Add all fields to the problem.
 * All fields can be added to the discrete problem via the constructor
 * of the base class \c SCADOPT::Problem.
 * - Evaluate all fields at all grid nodes.
 * All known fields can be evaluated at all global inner resp. border grid nodes
 * at all time steps by implementing the methods \c evalInner resp.
 * \c evalBorder.
 * - Initialise all unknown fields at all grid nodes.
 * All unknown fields can be initialised at all global inner resp. border
 * grid nodes at the starting time by implementing the methods \c initInner
 * resp. \c initBorder.
 * -  Update all unknown fields at all grid nodes.
 * All unknown fields can be updated at all global inner resp. border
 * grid nodes at the starting time by implementing the methods \c updateInner
 * resp. \c updateBorder.
 * -  Compute the target functional at all grid nodes.
 * The target functional can be computed at all global inner resp. border
 * grid nodes at a given timestep by implementing the methods \c compTfInner
 * resp. \c compTfBorder
 * - Write the main program.
 *
 * \section example A First Example
 * In the following, we show how a new PDE constrained optimization problem
 * can be implemented in SCADOPT. Fur this purpose, we have chosen a 3D
 * optimization problem based on the 3D heat equation on a cube as this example
 * serves as a prototype for the considered problem class.
 *
 * \ref optheatEqnFDM "Discretized optimization problem base on heat equation"
 *
 * \section contact Contact
 * Bugs can be reported to: kim.feldhoff@tu-dresden.de
 *
 * More information about SCADOPT like new releases can be found at:
 * <a href="http://www.tu-dresden.de/zih/scadopt">SCADOPT web page</a>
 *
 * \section funding Funding
 * SCADOPT was funded by the Federal Ministry of Education and Research (BMBF)
 * within the project HPC-FLiS under the support code 01 IH 11 009.
******************************************************************************/

#include "SCADOPT_Config.hpp"

#include "ScaFES.hpp"

#include "SCADOPT_Problem.hpp"

#endif
