/* SCADOPT
 * Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 * For details, see the files COPYING and LICENSE in the base directory
 * of the package.
 */

/**
 *  @file SCADOPT_Parameters.hpp
 *  @brief Contains the class Parameters.
 */

#ifndef SCADOPT_PARAMETERS_HPP_
#define SCADOPT_PARAMETERS_HPP_

#include "ScaFES_Config.hpp"

#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <vector>
#include <string>

#include <boost/serialization/serialization.hpp>

// #ifdef SCADOPT_HAVE_BOOST_SERIALIZATION
// namespace boost
// {
// namespace serialization
// {
// class access;
// }
// }
// #include <boost/serialization/vector.hpp>
// #include <boost/serialization/version.hpp>
// #endif


#ifdef SCADOPT_HAVE_BOOST_PROGRAM_OPTIONS
#include <boost/program_options.hpp>
#endif

#ifdef SCADOPT_HAVE_BOOST
#include <boost/lexical_cast.hpp>
#include <boost/format.hpp>
#endif

#ifdef SCADOPT_HAVE_BOOST_REGEX
#include <boost/regex.hpp>
#endif

#include "ScaFES_Ntuple.hpp"
#include "ScaFES_Communicator.hpp"

namespace SCADOPT
{

#ifdef SCADOPT_HAVE_BOOST_PROGRAM_OPTIONS
namespace po = boost::program_options;
#endif

/*******************************************************************************
 * GLOBAL VARIABLES.
 ******************************************************************************/
/** Maximal indentation depth level of console output. */
/*int gOutputLevelMax = 2;*/

/** Current indentation depth level of console output. */
/*int gOutputLevelCurr = gOutputLevelMax;*/

/*******************************************************************************
 ******************************************************************************/
/**
 * \class Parameters
 * @brief The class \c Parameters contains all parameters related to SCADOPT.
 *
 * Parameters can be given via the command line.
 * If a command line parameter is invalid, a help message will be displayed.
 *
 * The MPI environment lives as long as an instance of this class is living.
 * In particular, if the destructor of this class is called,
 * the MPI environment will be finalized. Thus, if the descructor is called
 * too early, the instance of the calls goes out of scope. So, please be careful
 * with calls to the destructor of this class.
 *
 * \remarks The master MPI process reads in the command line arguments
 * and distributed the parameters to all other MPI processes.
 */
class Parameters
{
public:
    /*----------------------------------------------------------------------
    | FRIEND CLASSES.
    ----------------------------------------------------------------------*/
    friend class boost::serialization::access;

    /*----------------------------------------------------------------------
    | LIFE CYCLE METHODS.
    ----------------------------------------------------------------------*/
    /** Creates own default constructor. */
    Parameters();

    /** Creates own constructor from command line parameters.
     * Parses the command line of a program call and provides a help
     * message.
     */
    Parameters(int argc, char* argv[]);

    /** Creates own copy constructor. */
    Parameters(Parameters const&);

    /** Creates own destructor. */
    ~Parameters();

    /** Creates own assignment operator. */
    Parameters& operator=(Parameters rhs);

    /*----------------------------------------------------------------------
    | GETTER METHODS.
     ----------------------------------------------------------------------*/
    /** Returns the space dimension. */
    int const& dim() const;

    /** Returns the number of grid nodes in each dimension. */
    std::vector<int> const& nNodes() const;

    /** Returns the coordinates of the first grid point. */
    std::vector<double> const& coordNodeFirst() const;

    /** Returns the coordinates of the last grid point. */
    std::vector<double> const& coordNodeLast() const;

    /** Returns the number of partitions in each direction. */
    std::vector<int> const& nPartitions() const;

    /** Returns the directions in which the grid should be partitioned. */
    std::vector<bool> const& divideGrid() const;

    /** Returns the decision if the kind field should be read in. */
    bool const& readKindFile() const;

    /** Returns if the kind field should be written to file. */
    bool const& writeKindFile() const;

    /** Returns the name of the kind file. */
    std::string const& kindFile() const;

    /** Returns the name of the domain decomposition approach. */
    std::string const& typeDomainDecomp() const;

    /** Returns the name of the partition file for writing. */
    std::string const& nameWritePartitionFile() const;

    /** Returns the decision if a partition file should be written. */
    bool const& writePartitionFile() const;

    /** Returns the name of the partition file for reading. */
    std::string const& nameReadPartitionFile() const;

    /** Returns the decision if a partition file should be written. */
    bool const& readPartitionFile() const;

    /** Returns the start of the time interval. */
    double const& timeIntervalStart() const;

    /** Returns the end of the time interval. */
    double const& timeIntervalEnd() const;

    /** Returns time step width. */
    double const& tau() const;

    /** Returns the number of time steps. */
    int const& nTimesteps() const;

    /** Returns the number of snapshots. */
    int const& nSnapshots() const;

    /** Returns the number of layers at the border. */
    int const& nLayersAtBorder() const;

    /** Returns the decision if an output file should be written. */
    bool const& writeDataFile() const;

    /** Returns the name of the output file. */
    std::string const& nameDataFile() const;

    /** Returns the name of the config file. */
    std::string const& nameConfigFile() const;

    /** Returns if ADOL-C is enabled. */
    bool const& enabledAdolc() const;

    /** Returns if asynchronous MPI communication is enabled. */
    bool const& asynchronMode() const;

    /** Returns if gradients should be computed. */
    bool const& computeGradients() const;

    /** Returns if the target functional is time-dependent. */
    bool const& isTfTimeDependent() const;

    /** Returns if gradients should be checked. */
    bool const& checkGradients() const;

    /** Returns the maximal number of iterations of the optimizer loop. */
    int const& nOptimIterationsMax() const;

    /** Returns the MPI rank for output. */
    int rankOutput() const;

    /** Returns the MPI communicator. */
    ScaFES::Communicator const& myWorld() const;

    /** Returns ScaFES parameters. */
    ScaFES::Parameters const& params() const;

    /** Returns the MPI rank. */
    int rank() const;

    /** Returns the maximum level of console output. */
    int const& indentDepth() const;

    /*----------------------------------------------------------------------
    | SETTER METHODS.
     ----------------------------------------------------------------------*/
    /** Increase output level by one. */
    void increaseLevel();

    /** Decrease output level by one. */
    void decreaseLevel();

    /*----------------------------------------------------------------------
    | COMPARISON METHODS.
    ----------------------------------------------------------------------*/
    /** Tests if two grids are equal. */
    bool operator==(Parameters const& rhs) const;

    /** Tests if two grids are not equal. */
    bool operator!=(Parameters const& rhs) const;

    /*----------------------------------------------------------------------
    | FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
    ----------------------------------------------------------------------*/
    /** Swaps to elements of this class. */
    friend void swap(Parameters& first, Parameters& second);

    /** Overloads the output operator \c operator<<. */
    friend std::ostream& operator<<(std::ostream& output, Parameters const& pp);

private:
    /*----------------------------------------------------------------------
    | WORK METHODS.
    ----------------------------------------------------------------------*/
    /** Checks a parameter given by its name. */
    template <typename TT>
    void checkParameter(po::variables_map const& vm,
                        std::string const& nameParam, std::vector<TT>& param,
                        bool isNecessary);

    /** Serializes class. */
    template <class Archive>
    void serialize(Archive& ar, const unsigned int version);

    /*----------------------------------------------------------------------
    | MEMBER VARIABLES.
    ----------------------------------------------------------------------*/
    /** Should gradients be computed? */
    bool mComputeGradients;

    /** Is the computatioon of the tf time-dependent? */
    bool mIsTfTimeDependent;

    /** Should gradients be checked? */
    bool mCheckGradients;

    /** Maximal number of iterations of the optimizer loop. */
    int mNoptimIterationsMax;

    /** Set of ScaFES parameters. */
    ScaFES::Parameters mParams;
}; // End of class. //

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
inline Parameters::Parameters()
: mComputeGradients(false)
, mIsTfTimeDependent(false)
, mCheckGradients(false)
, mNoptimIterationsMax(50)
, mParams()
{
}
/*----------------------------------------------------------------------------*/
inline Parameters::Parameters(int argc, char* argv[])
: mComputeGradients(false)
, mIsTfTimeDependent(false)
, mCheckGradients(false)
, mNoptimIterationsMax(50)
, mParams(argc, argv)
{
    if (0 == this->myWorld().rank())
    {

        po::variables_map vm;

        /*----------------------------------------------------------------------
        | Describe the options of the command line.
        ----------------------------------------------------------------------*/
        po::options_description desc("Allowed SCADOPT options:");
        desc.add_options()("help", "Displays this help message.")(
            "dim", po::value<int>(), "Sets the space dimension.")(
            "nPartitions", po::value<std::string>(),
            "Sets the number of partitions in each dimension (Np1x..xNpd).")(
            "typeDomainDecomposition", po::value<std::string>(),
            "Sets the type of domain decomposition: RCB (RCB algorithm), UNI "
            "(Uniform decomposition in each direction).")(
            "nNodes", po::value<std::string>(),
            "Sets the number of grid nodes in each dimension (N1x..xNd).")(
            "coordNodeFirst", po::value<std::string>(),
            "Sets the coordinates of the first grid node (c1x..xcd).")(
            "coordNodeLast", po::value<std::string>(),
            "Sets the coordinates of the last grid node (c1x..xcd).")(
            "divideGrid", po::value<std::string>(),
            "Divides the grid in which dimension? (g1x..xgd).")(
            "readKindFile", po::value<std::string>(),
            "Reads the kind field from a given kind file.")(
            "writeKindFile", po::value<int>(),
            "Writes the kind field to a NetCDF data file.")(
            "readPartitionfile", po::value<std::string>(),
            "Reads the partition table from a given partition file.")(
            "writePartitionfile", po::value<std::string>(),
            "Writes the partition table to a given partition file.")(
            "starttime", po::value<double>(), "Sets the start time.")(
            "endtime", po::value<double>(), "Sets the end time.")(
            "tau", po::value<double>(), "Sets the time step width.")(
            "nTimesteps", po::value<int>(),
            "Sets the number of time steps (=#{time intervals} (default=0).")(
            "nSnapshots", po::value<int>(), "Sets how many times the data "
                                            "fields are written to output "
                                            "(default=1).")(
            "nLayersAtBorder", po::value<int>(),
            "Sets of how many layers the border should consist (default=1).")(
            "outputfile", po::value<std::string>(),
            "Sets the name of the output data file.")(
            "configfile", po::value<std::string>(),
            "Sets the name of the config file.")(
            "enabledAdolc", po::value<int>(),
            "Sets if ADOL-C should be enabled (default=0)")(
            "asynchronMode", po::value<int>(), "Sets if asynchronous MPI "
                                               "communication should be used "
                                               "(default=1).")(
            "rankOutput", po::value<int>(),
            "Sets the rank of the MPI process whose output should be displayed "
            "(default=0).")("outputlevel", po::value<int>(),
                            "Controls the amount of console output.")(
            "computeGradients", po::value<int>(),
            "Sets if gradients should be computed (default=0).")(
            "isTfTimeDependent", po::value<int>(),
            "Sets if the target functional is time-dependent (default=0).")(
            "checkGradients", po::value<int>(), "Sets if gradients of forward "
                                                "computation should be checked "
                                                "(default=0).")(
            "nOptimIterationsMax", po::value<int>(), "Sets the maximal number "
                                                "of iterations of the optimizer"
                                                " loop (default=50). ");

        /*----------------------------------------------------------------------
        | Try to parse the command line.
        ----------------------------------------------------------------------*/
        try
        {
            po::store(po::parse_command_line(argc, argv, desc), vm);
            po::notify(vm);
        }
        catch (std::exception& e)
        {
            throw std::runtime_error("Wrong command line options.");
        }

        /*----------------------------------------------------------------------
        | Actions of command line options.
        ----------------------------------------------------------------------*/
        if (vm.count("help"))
        {
            std::cout << desc << "\n";
        }
        else {

        /*--------------------------------------------------------------------*/
        if (vm.count("computeGradients"))
        {
            int tmpComputeGradients = vm["computeGradients"].as<int>();
            if (0 == tmpComputeGradients)
            {
                this->mComputeGradients = false;
            }
            else
            {
                this->mComputeGradients = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --computeGradients=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("isTfTimeDependent"))
        {
            int tmpIsTfTimeDependent = vm["isTfTimeDependent"].as<int>();
            if (0 == tmpIsTfTimeDependent)
            {
                this->mIsTfTimeDependent = false;
            }
            else
            {
                this->mIsTfTimeDependent = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --isTfTimeDependent=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("checkGradients"))
        {
            int tmpCheckGradients = vm["checkGradients"].as<int>();
            if (0 == tmpCheckGradients)
            {
                this->mCheckGradients = false;
            }
            else
            {
                this->mCheckGradients = true;
            }
        }
        else
        {
            std::cerr << "\nREMARK: Use --checkGradients=<0/1>.\n\n";
        }
        /*--------------------------------------------------------------------*/
        if (vm.count("nOptimIterationsMax"))
        {
            this->mNoptimIterationsMax = vm["nOptimIterationsMax"].as<int>();
        }
        else
        {
            std::cerr << "\nREMARK: Use --nOptimIterationsMax=<int>.\n\n";
        }
        } // End of if block related to "help".
    }

    // Communicate result.
    ScaFES::broadcast(this->mParams.myWorld(), *this, 0);
}

/*----------------------------------------------------------------------------*/
inline Parameters::Parameters(Parameters const& rhs)
: mComputeGradients(rhs.computeGradients())
, mIsTfTimeDependent(rhs.isTfTimeDependent())
, mCheckGradients(rhs.checkGradients())
, mNoptimIterationsMax(rhs.nOptimIterationsMax())
, mParams(rhs.params())
{
}
/*----------------------------------------------------------------------------*/
inline Parameters& Parameters::operator=(Parameters rhs)
{
    swap(*this, rhs);
    return *this;
}
/*----------------------------------------------------------------------------*/
inline Parameters::~Parameters()
{
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
inline int const& Parameters::dim() const
{
    return this->params().dim();
}
/*----------------------------------------------------------------------------*/
inline std::vector<int> const& Parameters::nNodes() const
{
    return this->params().nNodes();
}
/*----------------------------------------------------------------------------*/
inline std::vector<double> const& Parameters::coordNodeFirst() const
{
    return this->params().coordNodeFirst();
}
/*----------------------------------------------------------------------------*/
inline std::vector<double> const& Parameters::coordNodeLast() const
{
    return this->params().coordNodeLast();
}
/*----------------------------------------------------------------------------*/
inline std::vector<int> const& Parameters::nPartitions() const
{
    return this->params().nPartitions();
}
/*----------------------------------------------------------------------------*/
inline std::vector<bool> const& Parameters::divideGrid() const
{
    return this->params().divideGrid();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::readKindFile() const
{
    return this->params().readKindFile();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writeKindFile() const
{
    return this->params().writeKindFile();
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::kindFile() const
{
    return this->params().kindFile();
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::typeDomainDecomp() const
{
    return this->params().typeDomainDecomp();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writePartitionFile() const
{
    return this->params().writePartitionFile();
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameWritePartitionFile() const
{
    return this->params().nameWritePartitionFile();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::readPartitionFile() const
{
    return this->params().readPartitionFile();
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameReadPartitionFile() const
{
    return this->params().nameReadPartitionFile();
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::timeIntervalStart() const
{
    return this->params().timeIntervalStart();
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::timeIntervalEnd() const
{
    return this->params().timeIntervalEnd();
}
/*----------------------------------------------------------------------------*/
inline double const& Parameters::tau() const
{
    return this->params().tau();
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nTimesteps() const
{
    return this->params().nTimesteps();
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nSnapshots() const
{
    return this->params().nSnapshots();
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nLayersAtBorder() const
{
    return this->params().nLayersAtBorder();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::writeDataFile() const
{
    return this->params().writeDataFile();
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameDataFile() const
{
    return this->params().nameDataFile();
}
/*----------------------------------------------------------------------------*/
inline std::string const& Parameters::nameConfigFile() const
{
    return this->params().nameConfigFile();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::enabledAdolc() const
{
    return this->params().enabledAdolc();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::asynchronMode() const
{
    return this->params().asynchronMode();
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::computeGradients() const
{
    return this->mComputeGradients;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::isTfTimeDependent() const
{
    return this->mIsTfTimeDependent;
}
/*----------------------------------------------------------------------------*/
inline bool const& Parameters::checkGradients() const
{
    return this->mCheckGradients;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::nOptimIterationsMax() const
{
    return this->mNoptimIterationsMax;
}
/*----------------------------------------------------------------------------*/
inline int Parameters::rankOutput() const
{
    return this->params().rankOutput();
}
/*----------------------------------------------------------------------------*/
inline ScaFES::Communicator const& Parameters::myWorld() const
{
    return this->mParams.myWorld();
}
/*----------------------------------------------------------------------------*/
inline ScaFES::Parameters const& Parameters::params() const
{
    return this->mParams;
}
/*----------------------------------------------------------------------------*/
inline int const& Parameters::indentDepth() const
{
    return this->params().indentDepth();
}

/*******************************************************************************
 * SETTER METHODS.
 ******************************************************************************/
inline void Parameters::increaseLevel()
{
    ++(ScaFES::gOutputLevelCurr);
}
/*----------------------------------------------------------------------------*/
inline void Parameters::decreaseLevel()
{
    --(ScaFES::gOutputLevelCurr);
}

/*******************************************************************************
 * COMPARISON METHODS.
 ******************************************************************************/
inline bool Parameters::operator==(Parameters const& rhs) const
{
    bool isEqual = true;

    if (this->computeGradients() != rhs.computeGradients())
    {
        isEqual = false;
    }
    if (this->isTfTimeDependent() != rhs.isTfTimeDependent())
    {
        isEqual = false;
    }
    if (this->checkGradients() != rhs.checkGradients())
    {
        isEqual = false;
    }
    if (this->nOptimIterationsMax() != rhs.nOptimIterationsMax())
    {
        isEqual = false;
    }
    if (this->params() != rhs.params())
    {
        isEqual = false;
    }

    return isEqual;
}
/*----------------------------------------------------------------------------*/
inline bool Parameters::operator!=(Parameters const& rhs) const
{
    return !(*this == rhs);
}

/*----------------------------------------------------------------------------*/
template <class Archive>
void Parameters::serialize(Archive& ar, const unsigned int version)
{
    if (1 <= version)
    {
        ar&(this->mComputeGradients);
        ar&(this->mIsTfTimeDependent);
        ar&(this->mCheckGradients);
        ar&(this->mNoptimIterationsMax);
        ar&(this->mParams);
    }
}

/*******************************************************************************
 * FREE METHODS WHICH ARE FRIENDS OF THIS CLASS.
 ******************************************************************************/
inline void swap(Parameters& first, Parameters& second)
{
    std::swap(first.mComputeGradients, second.mComputeGradients);
    std::swap(first.mIsTfTimeDependent, second.mIsTfTimeDependent);
    std::swap(first.mCheckGradients, second.mCheckGradients);
    std::swap(first.mNoptimIterationsMax, second.mNoptimIterationsMax);
    std::swap(first.mParams, second.mParams);
}

/*******************************************************************************
 * FREE FUNCTIONS.
 ******************************************************************************/
/** Free function which sets a prefix to the output of process with
 * identifier \c rank. */
inline std::string getPrefix(int const& indentationDepth, int const& rank,
                             std::string const& nameFramework =
                                 std::string("SCADOPT "))
{
    return ScaFES::getPrefix(indentationDepth, rank, nameFramework);
}

} // End of namespace. //

/*******************************************************************************
 ******************************************************************************/
/** Boost serialization version of class \c Parameter. */
BOOST_CLASS_VERSION(SCADOPT::Parameters, 2);

#endif
