/* SCADOPT
 * Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
 * For details, see the files COPYING and LICENSE in the base directory
 * of the package.
 */

/**
 *  @file SCADOPT_Problem.hpp
 *  @brief Contains the class template Problem.
 */

#ifndef SCADOPT_PROBLEM_HPP_
#define SCADOPT_PROBLEM_HPP_

#include "SCADOPT_Config.hpp"

#include <iostream>
#include <ios>
#include <fstream>
#include <string>
#include <vector>
#include <tuple>
#include <cassert>
#include <unordered_map>

extern "C" {
#ifdef _OPENMP
#include <omp.h>
#endif
}

#ifdef SCADOPT_HAVE_ADOLC
#include <adolc/adolc.h>
#include <adolc/adouble.h>
#ifdef _OPENMP
#include <adolc/adolc_openmp.h>
#endif
#endif

#ifdef SCADOPT_HAVE_BOOST
#include <boost/type_traits.hpp>
#include <boost/assign/list_of.hpp>
#endif

#ifdef VTRACE
#include "vt_user.h"
#endif

#ifdef SCADOPT_HAVE_LBFGSB
extern "C" {
#include <lbfgsb.h>
}
#endif

#include "ScaFES_Timer.hpp"
#include "ScaFES.hpp"

#include "SCADOPT_Parameters.hpp"

/** \namespace SCADOPT
 * \brief Namespace of the SCADOPT package. */
namespace SCADOPT
{

/*******************************************************************************
 ******************************************************************************/
/**
 * \class Problem
 * @brief Base class for PDE constrained optimization problems.
 * The class \c Problem is designed to describe a PDE constrained
 * optimization problem on a given grid partition. The optimization problem
 * usually consists of an initial boundary value problem, a target functional,
 * and an initial guess of the (geometrical) parameters.
 *
 * The user has to implement an own class that is inherited by this class.
 * The derived class has to contain the following methods:
 * - \c evalInner
 * - \c evalBorder
 * - \c initInner
 * - \c initBorder
 * - \c updateInner
 * - \c updateBorder
 * - \c compTfInner
 * - \c compTfBorder
 *
 * If the leap frog scheme is used then the user has to provide additionally
 * the following methods:
 * - \c updateInner2
 * - \c updateBorder2
 *
 * These methods must be declared public in the sub class
 * as they will be accessed within in this class via the CRTP.
 */
template <class OWNPRBLM, typename CT = double, std::size_t DIM = 3>
class Problem : public ScaFES::Problem<Problem<OWNPRBLM, CT, DIM>, CT, DIM>
{
public:
    /*--------------------------------------------------------------------------
    | TYPE DEFINITIONS + ALIAS DECLARATIONS (C++11).
    --------------------------------------------------------------------------*/
    /** Re-export typename CT STL-like. */
    typedef CT value_type;

    /** Alias declaration for vector of data fields. */
    template <typename TT>
    using SCADOPT_VectDf = std::vector<ScaFES::DataField<TT, DIM>>;

    /** Type definition for integer tuple. */
    typedef ScaFES::Ntuple<int, DIM> SCADOPT_IntNtuple;

    /*--------------------------------------------------------------------------
    | LIFE CYCLE METHODS.
    --------------------------------------------------------------------------*/
    /** Forbids default constructor. */
    Problem() = delete;

    /** Creates own constructor.
     * All data fields which are related to the underlying problem
     * have to be added in terms of an entry of the parameters of
     * type \c std::vector.
     * @param params Set of SCADOPT parameters.
     * @param gg Global grid.
     * @param useLeapfrog Should the leap frog scheme be used?
     * @param nameDatafield Name of the data fields.
     * @param stencilWidth Stencil width of the data fields.
     * @param isKnownDf Is the data field are known or unknown one?
     * @param nLayers Number of layers at the global boundary.
     * @param optimizeDf After which data fields should be optimized?
     * @param defaultValue Default value of data fields.
     * @param writeToFile How often should the data field be written
     *                    to file.
     * @param computeError Should the Linf error between the numerical
     *                     and exact solution be computed?
     * @param geomparamsInit Initial guess of geometrical parameters.
     * @param geomparamsSol Exact solution of geometrical parameters
     *                      (if known).
     */
    Problem(SCADOPT::Parameters const& params, ScaFES::GridGlobal<DIM> const& gg,
            bool const& useLeapfrog,
            std::vector<std::string> const& nameDatafield,
            std::vector<int> const& stencilWidth,
            std::vector<bool> const& isKnownDf,
            std::vector<int> const& nLayers = std::vector<int>(),
            std::vector<bool> const& optimizeDf = std::vector<bool>(),
            std::vector<CT> const& defaultValue = std::vector<CT>(),
            std::vector<ScaFES::WriteHowOften> const& writeToFile =
                std::vector<ScaFES::WriteHowOften>(),
            std::vector<bool> const& computeError = std::vector<bool>(),
            std::vector<CT> const& geomparamsInit = std::vector<CT>(),
            std::vector<CT> const& geomparamsSol = std::vector<CT>());

    /** Creates compiler provided constructor. */
    Problem(Problem<OWNPRBLM, CT, DIM> const& /*ppp*/) = default;

    /** Creates compiler provided destructor. */
    virtual ~Problem() = default;

    /** Creates compiler provided assignment operator. */
    Problem& operator=(Problem<OWNPRBLM, CT, DIM> const& /*ppp*/) = default;

    /*--------------------------------------------------------------------------
    | GETTER METHODS.
    --------------------------------------------------------------------------*/
    /** Returns current value of target functional. */
    CT const& tf() const;

    /** Returns current value of gradient of target functional. */
    std::vector<CT> const& gradTf() const;

    /** Returns the set of SCADOPT parameters. */
    SCADOPT::Parameters const& params() const;

    /*--------------------------------------------------------------------------
    | WORK METHODS.
    --------------------------------------------------------------------------*/
    /** Computes new target functional and its gradient. */
    void compNewTfAndGradTf();

    /** Iterates over time (including computation of tf). */
    void iterateOverTimeInternal();

    /** Solves the current inverse problem by optimizing a target function
     * using a given optimizer. Currently: "lbfgsb".
     * This is the interface for optimizers like L-BFGS-B or IP-OPT. */
    void minimizeTargetfct(std::string const& optimizer);

    /** Evaluates ALL data fields at one given global inner grid node.
     *  @param dfDep Set of all data fields.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void evalInner(SCADOPT_VectDf<TT>& dfDep, SCADOPT_IntNtuple const& idxNode,
                   int const& timeIter);

    /** Evaluates ALL data fields at one given global border grid node.
     *  @param dfDep Set of all data fields.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void evalBorder(SCADOPT_VectDf<TT>& dfDep, SCADOPT_IntNtuple const& idxNode,
                    int const& timeIter);

    /** Initializes all unknown data fields at one given global inner
     *  grid node.
     *  @param dfDep Set of all unknown data fields.
     *  @param dfIndep Set of all geometrical parameters.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void initInner(SCADOPT_VectDf<TT>& dfDep, std::vector<TT> const& dfIndep,
                   SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Initializes all unknown data fields at one given global border
     *  grid node.
     *  @param dfDep Set of all unknown data fields.
     *  @param dfIndep Set of all geometrical parameters.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void initBorder(SCADOPT_VectDf<TT>& dfDep, std::vector<TT> const& dfIndep,
                    SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Initializes all unknown data fields at one given global inner
     *  grid node.
     *  @param dfDep Set of all unknown data fields.
     *  @param dfIndep Set of all given data fields.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void initInner(SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
                   SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Initializes all unknown data fields at one given global border
     *  grid node.
     *  @param dfDep Set of all unknown data fields.
     *  @param dfIndep Set of all given data fields.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void initBorder(SCADOPT_VectDf<TT>& dfDep,
                    SCADOPT_VectDf<TT> const& dfIndep,
                    SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Updates all unknown data fields at one given global inner
     *  grid node.
     *  @param dfDep Set of all unknown data fields at new time step.
     *  @param dfIndep Set of all unknown data fields at old time step.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void updateInner(SCADOPT_VectDf<TT>& dfDep,
                     SCADOPT_VectDf<TT> const& dfIndep,
                     SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Updates all unknown data fields at one given global border
     *  grid node.
     *  @param dfDep Set of all unknown data fields at new time step.
     *  @param dfIndep Set of all unknown data fields at old time step.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void updateBorder(SCADOPT_VectDf<TT>& dfDep,
                      SCADOPT_VectDf<TT> const& dfIndep,
                      SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Updates (2nd cycle) all unknown data fields at one given global
     *  inner grid node using the leap frog scheme.
     *  @param dfDep Set of all unknown data fields at new time step.
     *  @param dfIndep Set of all unknown data fields at old time step.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void updateInner2(SCADOPT_VectDf<TT>& dfDep,
                      SCADOPT_VectDf<TT> const& dfIndep,
                      SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Updates (2nd cycle) all unknown data fields at one given global
     *  border grid node using the leap frog scheme.
     *  @param dfDep Set of all unknown data fields at new time step.
     *  @param dfIndep Set of all unknown data fields at old time step.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void updateBorder2(SCADOPT_VectDf<TT>& dfDep,
                       SCADOPT_VectDf<TT> const& dfIndep,
                       SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Computes target functional value at one given global
     *  inner grid node using the leap frog scheme.
     *  @param dfDep Target functional value.
     *  @param dfIndep Set of all unknown data fields at current time step.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void compTfInner(TT& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
                     SCADOPT_IntNtuple const& idxNode, int const& timeIter);

    /** Computes target functional value at one given global
     *  border grid node using the leap frog scheme.
     *  @param dfDep Target functional value.
     *  @param dfIndep Set of all unknown data fields at current time step.
     *  @param idxNode Index of given grid node.
     *  @param timeIter Given time iterate.
     * \remark The concrete function must be implemented
     * in the user provided sub class.
     */
    template <typename TT>
    void compTfBorder(TT& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
                      SCADOPT_IntNtuple const& idxNode, int const& timeIter);

private:
    /***************************************************************************
    * CLASS.
    ***************************************************************************/
    class LbfgsbParams
    {
    public:
        LbfgsbParams() = default;
        void initLbfgsbParams(
            int const& nParamsToOptimize,
            int const& nCorrectionsInMatrix,
            int const& optimVersion
        )
        {
            this->scalarTf = 0.0;
            this->nCorrectionsInMatrix = nCorrectionsInMatrix;
            this->memoryIterate.resize(nParamsToOptimize, 0.0);
            this->memoryGradTf.resize(nParamsToOptimize, 0.0);
            this->lowerBound.resize(nParamsToOptimize, 0.0);
            this->upperBound.resize(nParamsToOptimize, 0.0);
            // All variables are unbounded.
            this->typeOfBounds.resize(nParamsToOptimize, 0);

            this->factr = 1.0e1;
            this->pgtol = 1.0e-12; // Criterium will be ignored.
            const int nmax = nParamsToOptimize;
            const int mmax = nCorrectionsInMatrix;
            info = 0;

            int nWa = 0;
            if (2 < optimVersion)
            {
                // L-BFGS-B 3.0
                nWa = (2 * mmax + 5) * nmax + 11 * mmax * mmax + 8 * mmax;
            }
            else
            {
                // L-BFGS-B 2.1
                nWa = (2 * mmax * nmax + 4 * nmax
                       + 12 * mmax * mmax + 12 * mmax);
            }
            this->wa.resize(nWa, 0.0);
            this->iwa.resize(3 * nmax, 0);
            this->csave.resize(60, static_cast<char>(0));
            this->lsave.resize(4, 0);
            this->isave.resize(44, 0);
            this->dsave.resize(29, 0.0);
            for (int ii = 0; ii < 60; ++ii)
            {
                task[ii] = static_cast<char>(0);
            }
            task[59] = '\0';

            task[0] = 'S';
            task[1] = 'T';
            task[2] = 'A';
            task[3] = 'R';
            task[4] = 'T';
        }

    public:
        double scalarTf;
        std::vector<double> memoryIterate;
        std::vector<double> memoryGradTf;
        int nCorrectionsInMatrix;
        std::vector<double> lowerBound;
        std::vector<double> upperBound;
        std::vector<int> typeOfBounds;
        double factr;
        double pgtol;
        int info;
        std::vector<double> wa;
        std::vector<int> iwa;
        std::vector<char> csave;
        std::vector<int> lsave;
        std::vector<int> isave;
        std::vector<double> dsave;
        char task[60];
    };

    /*--------------------------------------------------------------------------
    | GETTER METHODS.
    --------------------------------------------------------------------------*/
    /** Returns if the target functional phase was already traced. */
    bool const& isTracedTfPhase() const;

    /** Returns the tape identifier corresponding to the tape containing
     *  the target functional at all inner partition grid nodes. */
    short int const& tapeIdComputeTfPartitionInner() const;

    /** Returns the tape identifier corresponding to the tape containing
     *  the target functional at all border partition grid nodes. */
    short int const& tapeIdComputeTfPartitionBorder() const;

    /** Returns the tape identifier corresponding to the tape containing
     *  the target functional at all partition grid nodes. */
    short int const& tapeIdComputeTfPartitionAll() const;

    /** Returns the number of variables during optimization (memNormal). */
    int const& nParamsToOptimize() const;

    /** Returns the number of directions. */
    std::vector<bool> const& optimizeDf() const;

    /*--------------------------------------------------------------------------
    | WORK METHODS.
    --------------------------------------------------------------------------*/
    /** Solves the current inverse problem by optimizing a target function
     * using L-BFGS-B and based on the grid ansatz or geometrical parameters
     * ansatz. */
    void minimizeTargetfctUsingLbfgsb(int const& optimVersion);

    /*------------------------------------------------------------------------*/
    /** Evaluates: Computes the value of the target functional.
     *  The computation is included in a loop over all time steps
     *  as the target functional can be time-dependent. */
    void evalTfPhase(int const& timeIter);

    /** Traces the target functional. */
    void traceTfPhase(int const& timeIter);

    /*------------------------------------------------------------------------*/
    /** Computes a vector containing the number of elements from each data field
     * which should be optimized. */
    void compVectOfNelemsToOptimize(
        std::vector<int>& nParamsToOptimizeLocalOfDf,
        int const& nDfsToOptimize
    );

    /** Gets the number of data fields after which should be optimized. */
    int getNdataFieldsForOptimization() const;

    /*--------------------------------------------------------------------------
    | MEMBER VARIABLES.
    --------------------------------------------------------------------------*/
    /* Set of SCADOPT parameters. */
    SCADOPT::Parameters mParams;

    /** Vector of Wall clock times: Contains accumulated times
     * of all forward iterations.
     * tIterate, tTrace, tInit, tUpdate, tCompTf,
     * tSync, tWrite, tSolution. */
    std::unordered_map<std::string, double> mWallClockTimesOptim;

    /** Tape identifier of target functional phase. */
    short int mTapeIdComputeTfPartitionInner;

    /** Tape identifier of target functional phase. */
    short int mTapeIdComputeTfPartitionBorder;

    /** Tape identifier of target functional phase. */
    short int mTapeIdComputeTfPartitionAll;

    /** Is target functional already traced? */
    bool mIsTracedTfPhase;

    /** Optimize data field? */
    std::vector<bool> mOptimizeDf;

    /** Geomparams based ansatz: Reference if known. */
    std::vector<CT> mVectParamsRef;

    /** Value of target functional. */
    CT mTf;

    /** Grid based ansatz: Gradient of target functional (at memAll). */
    SCADOPT_VectDf<CT> mVectGradTf;

    /** Memory for gradient of target functional.
     * This will be used for the grid based AND the parameter based ansatz.
     * \remark Has to be a continuous memory lump for ADOL-C and L-BFGS-B.
     */
    std::vector<CT> mMemoryVectGradTf;

    /** Number of variables which should be optimized (memNormal). */
    int mNparamsToOptimize;

    /** Iterate of optimization loop (at memNormal without halo nodes). */
    SCADOPT_VectDf<CT> mVectIterate;

    /** Memory for storing values of iterate of optimization loop. */
    std::vector<CT> mMemoryVectIterate;

    /** Gradient of target functional
     * within optimization loop (at memNormal without halo nodes). */
    SCADOPT_VectDf<CT> mVectIterateGradTf;

    /** Memory for storing values of gradient of tf of optimization loop. */
    std::vector<CT> mMemoryVectIterateGradTf;
}; // End of class. //

/*******************************************************************************
 * LIFE CYCLE METHODS.
 ******************************************************************************/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline Problem<OWNPRBLM, CT, DIM>::Problem(
    SCADOPT::Parameters const& params, ScaFES::GridGlobal<DIM> const& gg,
    bool const& useLeapfrog, std::vector<std::string> const& nameDatafield,
    std::vector<int> const& stencilWidth, std::vector<bool> const& isKnownDf,
    std::vector<int> const& nLayers, std::vector<bool> const& optimizeDf,
    std::vector<CT> const& defaultValue,
    std::vector<ScaFES::WriteHowOften> const& writeToFile,
    std::vector<bool> const& computeError,
    std::vector<CT> const& geomparamsInit, std::vector<CT> const& geomparamsSol)
: ScaFES::Problem<Problem<OWNPRBLM, CT, DIM>, CT, DIM>(
      params.params(), gg, useLeapfrog, nameDatafield, stencilWidth,
                                                       isKnownDf, nLayers,
      defaultValue, writeToFile, computeError, geomparamsInit),
  mParams(params),
  mWallClockTimesOptim(),
  mTapeIdComputeTfPartitionInner(21844 + this->myRank()),
  mTapeIdComputeTfPartitionBorder(22844 + this->myRank()),
  mTapeIdComputeTfPartitionAll(23844 + this->myRank()), mIsTracedTfPhase(false),
  mOptimizeDf(optimizeDf), mVectParamsRef(geomparamsSol),
  mTf(static_cast<CT>(0)), mVectGradTf(), mMemoryVectGradTf(),
  mNparamsToOptimize(0), mVectIterate(), mMemoryVectIterate(),
  mVectIterateGradTf(), mMemoryVectIterateGradTf()
{
    this->mWallClockTimesOptim["communic"] = 0.0;
    this->mWallClockTimesOptim["lbfgsb"] = 0.0;
    this->mWallClockTimesOptim["compTf"] = 0.0;

    /*------------------------------------------------------------------------*/
    int idDf = 0;
    int idOptimizeAndUnknownDf = 0;
    int nDataFields = nameDatafield.size();

    /*------------------------------------------------------------------------*/
    int nNodesTotalAllUnknownDf = 0;
    for (std::size_t ii = 0; ii < this->vectUnknownDfsDomNew().size(); ++ii)
    {
        nNodesTotalAllUnknownDf +=
            this->vectUnknownDfsDomNew()[ii].nElemsAllocated();
    }

    /*------------------------------------------------------------------------*/
    // TODO: Check if optimizer should work parameter based or grid based.
    // this->mIsParamsBased = ...

    /*------------------------------------------------------------------------*/
    if (!(this->isParamsBased()))
    {
        int idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!this->isKnownDf().at(ii))
            {
                if (this->optimizeDf().at(ii))
                {
                    this->mNparams +=
                        this->vectUnknownDfsDomNew().at(idDf).nElemsAllocated();
                }
                ++idDf;
            }
        }
    }
    /*------------------------------------------------------------------------*/
    // Create memory for local arrays.
    this->mNparamsToOptimize = 0;
    if (this->isParamsBased())
    {
        this->mNparamsToOptimize = this->vectParamsCurr().size();
    }
    else
    {
        this->mNparamsToOptimize = 0;
        idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!this->isKnownDf().at(ii))
            {
                if (this->optimizeDf().at(ii))
                {
                    this->mNparamsToOptimize += this->vectUnknownDfsDomNew()
                                                    .at(idDf)
                                                    .memNormal()
                                                    .nNodesSub()
                                                    .size();
                }
                ++idDf;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of unknown data fields.
    int nUnknownDfs = 0;
    for (int ii = 0; ii < nDataFields; ++ii)
    {
        if (!this->isKnownDf().at(ii))
        {
            ++nUnknownDfs;
        }
    }

    /*------------------------------------------------------------------------*/
    // Determine number of elements (memAll) of known and unknown data fields.
    std::vector<SCADOPT_IntNtuple> nMemAllOfDf(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii)
    {
        SCADOPT_IntNtuple idxNodeFirst =
            gg.partition(params.myWorld().rank()).idxNodeFirstSub() -
            SCADOPT_IntNtuple(stencilWidth.at(ii));
        SCADOPT_IntNtuple idxNodeLast =
            gg.partition(params.myWorld().rank()).idxNodeLastSub() +
            SCADOPT_IntNtuple(stencilWidth.at(ii));
        nMemAllOfDf[ii] = idxNodeLast - idxNodeFirst + SCADOPT_IntNtuple(1);
    }

    /*------------------------------------------------------------------------*/
    // Determine number of elements (memNormal) of known + unknown data fields.
    int nMemNormalUnknownDfs = 0;
    int nMemNormalUnknownAndOptimizeDfs = 0;
    std::vector<SCADOPT_IntNtuple> nMemNormalOfDf(nDataFields);
    for (int ii = 0; ii < nDataFields; ++ii)
    {
        SCADOPT_IntNtuple idxNodeFirstNormal =
            gg.partition(params.myWorld().rank()).idxNodeFirstSub();
        SCADOPT_IntNtuple idxNodeLastNormal =
            gg.partition(params.myWorld().rank()).idxNodeLastSub();
        nMemNormalOfDf[ii] =
            idxNodeLastNormal - idxNodeFirstNormal + SCADOPT_IntNtuple(1);
        if (!this->isKnownDf().at(ii))
        {
            nMemNormalUnknownDfs += nMemNormalOfDf[ii].size();
            if (this->optimizeDf().at(ii))
            {
                nMemNormalUnknownAndOptimizeDfs += nMemNormalOfDf[ii].size();
            }
        }
    }


    /*------------------------------------------------------------------------*/
    const int NCOMPS_TF = 1;

    // PARAMSBASED ANSATZ.
    if (this->isParamsBased())
    {
        // Create memory for gradient of target functional (memAll)
        // Create data field for gradient of tf.
        if (0 < this->nColumns())
        {
            this->mMemoryVectGradTf.resize(NCOMPS_TF * this->nColumns(),
                                       static_cast<CT>(0));
        }
        if (0 < this->nParamsToOptimize())
        {
           this->mMemoryVectIterate.resize(this->nParamsToOptimize(),
                                        static_cast<CT>(0));
        }
        if (0 < this->nParamsToOptimize())
        {
            this->mMemoryVectIterateGradTf.resize(this->nParamsToOptimize(),
                                                  static_cast<CT>(0));
        }
    }


    /*------------------------------------------------------------------------*/
    // GRIDBASED ANSATZ.
    // Create memory for gradient of target functional (memAll)
    // Create data field for gradient of tf.
    if (!(this->isParamsBased()))
    {
        if (0 < this->nColumns())
        {
            this->mMemoryVectGradTf.resize(NCOMPS_TF * this->nColumns(),
                                       static_cast<CT>(0));
        }

        // Create corresponding data field only in case of the grid based ansatz.
        int nDataFieldsForOptimization = this->getNdataFieldsForOptimization();
        CT* elemDataCurr = this->mMemoryVectGradTf.data();
        this->mVectGradTf.reserve(nDataFieldsForOptimization);
        idDf = 0;
        idOptimizeAndUnknownDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!this->isKnownDf().at(ii))
            {
                if (this->optimizeDf().at(ii))
                {
                    std::string nameDatafieldCurr =
                        nameDatafield.at(ii) + "GradTf";
                    this->mVectGradTf.push_back(ScaFES::DataField<CT, DIM>(
                        nameDatafieldCurr, this->params().params(),
                        this->globalGrid(),
                        this->vectUnknownDfsDomOld()[idDf].memAll(),
                        stencilWidth.at(ii), elemDataCurr, NCOMPS_TF,
                        this->vectUnknownDfsDomOld()[idDf].memHole(),
                        nLayers.at(ii), writeToFile.at(ii)));
                    elemDataCurr += (nMemAllOfDf.at(ii).size() * NCOMPS_TF);
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
        /*--------------------------------------------------------------------*/
        // Set default value for each iterate data field.
        idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!(this->isKnownDf().at(ii)))
            {
                if (this->optimizeDf().at(ii))
                {
                    this->mVectGradTf.at(idDf) = defaultValue.at(ii);
                    ++idDf;
                }
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Create (local) iterate of optimization algorithm.
    // Do not create a corresponding data field for the (local) iterate
    // in case of the parameter based ansatz.
    if (! (this->isParamsBased()))
    {
        if (0 < nMemNormalUnknownAndOptimizeDfs)
        {
            this->mMemoryVectIterate.resize(nMemNormalUnknownAndOptimizeDfs * 1,
                                            static_cast<CT>(0));
        }
        int nDataFieldsForOptimization = this->getNdataFieldsForOptimization();
        CT* elemDataCurr = this->mMemoryVectIterate.data();
        this->mVectIterate.reserve(nDataFieldsForOptimization);
        idDf = 0;
        idOptimizeAndUnknownDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!this->isKnownDf().at(ii))
            {
                if (this->optimizeDf().at(ii))
                {
                    std::string nameDatafieldOld =
                        nameDatafield.at(ii) + "Iterate";
                    this->mVectIterate.push_back(ScaFES::DataField<CT, DIM>(
                        nameDatafieldOld, this->params().params(),
                        this->globalGrid(),
                        this->vectUnknownDfsDomOld()[idDf].memAll(), 0,
                        elemDataCurr, 1,
                        this->vectUnknownDfsDomOld()[idDf].memHole(),
                        nLayers.at(ii), writeToFile.at(ii)));
                    elemDataCurr += nMemNormalOfDf.at(ii).size() * 1;
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }

        /*--------------------------------------------------------------------*/
        // Set default value for each iterate data field.
        idDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!(this->isKnownDf().at(ii)))
            {
                if (this->optimizeDf().at(ii))
                {
                    this->mVectIterate.at(idDf) = defaultValue.at(ii);
                    ++idDf;
                }
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // Create gradient of target func. for optimization algorithm (memNormal).
    if (!(this->isParamsBased()))
    {
        if (0 < nMemNormalUnknownAndOptimizeDfs)
        {
            this->mMemoryVectIterateGradTf.resize(
                nMemNormalUnknownAndOptimizeDfs * 1, static_cast<CT>(0));
        }
        int nDataFieldsForOptimization = this->getNdataFieldsForOptimization();
        CT* elemDataCurr = this->mMemoryVectIterateGradTf.data();
        this->mVectIterateGradTf.reserve(nDataFieldsForOptimization);
        idDf = 0;
        idOptimizeAndUnknownDf = 0;
        for (int ii = 0; ii < nDataFields; ++ii)
        {
            if (!(this->isKnownDf().at(ii)))
            {
                if (this->optimizeDf().at(ii))
                {
                    std::string nameDatafieldCurr =
                        nameDatafield.at(ii) + "GradTfIter";
                    this->mVectIterateGradTf.push_back(
                        ScaFES::DataField<CT, DIM>(
                            nameDatafieldCurr, this->params().params(),
                            this->globalGrid(),
                            this->vectUnknownDfsDomOld()[idDf].memAll(), 0,
                            elemDataCurr, 1,
                            this->vectUnknownDfsDomOld()[idDf].memHole(),
                            nLayers.at(ii)));
                    elemDataCurr += nMemNormalOfDf.at(ii).size() * 1;
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
    }
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline CT const& Problem<OWNPRBLM, CT, DIM>::tf() const
{
    return this->mTf;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<CT> const& Problem<OWNPRBLM, CT, DIM>::gradTf() const
{
    return this->mMemoryVectGradTf;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline SCADOPT::Parameters const& Problem<OWNPRBLM, CT, DIM>::params() const
{
    return this->mParams;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::compNewTfAndGradTf()
{
    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Compute new tf and grad(tf)..." << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // Reset target functional value and its gradient to zero.
    this->mTf = static_cast<CT>(0);
    if (this->isParamsBased())
    {
        for (std::size_t ll = 0; ll < this->gradTf().size(); ++ll)
        {
            this->mMemoryVectGradTf[ll] = static_cast<CT>(0);
        }
    }
    /*------------------------------------------------------------------------*/
    // Copy initial guess to first iterate of optimization loop.
    if (this->isParamsBased())
    {
        for (int ll = 0; ll < this->nParamsToOptimize(); ++ll)
        {
            this->mMemoryVectIterate[ll] = this->vectParamsCurr().at(ll);
        }
    }
    // REMARK: Initial guess is already given as MPI local data
    // field mVectUnknownDfsDomOld.
    // * mVectIterate works on grid of memNormal.
    // * mVectUnknownDfsDomOld works on grid of memAll.
    else
    {
        int idOptimizeAndUnknownDf = 0;
        int idDf = 0;
        for (std::size_t ii = 0; ii < this->optimizeDf().size(); ++ii)
        {
            if (!this->isKnownDf().at(ii))
            {
                if (this->optimizeDf().at(ii))
                {
                    this->mVectIterate[idOptimizeAndUnknownDf]
                        .setValuesAtMemNormal(
                             this->vectUnknownDfsDomOld()[idDf]);
                    ++idOptimizeAndUnknownDf;
                }
                ++idDf;
            }
        }
        // REMARK: Synchronization of mVectIterate is not necessary
        // because this data field only lives on memNormal!
    }

    this->iterateOverTimeInternal();

    /*------------------------------------------------------------------------*/
    // Copy resp. convert gradient of target functional.
    if (this->isParamsBased())
    {
        for (int ll = 0; ll < this->nParamsToOptimize(); ++ll)
        {
            this->mMemoryVectIterateGradTf[ll] = this->mMemoryVectGradTf[ll];
        }
    }
    else
    {
        for (std::size_t ii = 0; ii < this->mVectGradTf.size(); ++ii)
        {
            this->mVectGradTf[ii].sync(this->nTimesteps());
        }

        // Both objects have the same vector lengths.
        for (std::size_t ii = 0; ii < this->mVectGradTf.size(); ++ii)
        {
            this->mVectIterateGradTf[ii].setValuesAtMemNormal(mVectGradTf[ii]);
            // mVectIterateGradTf need not to be synchronized as it lives
            // on memNormal.
        }
    }

    /*------------------------------------------------------------------------*/
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::iterateOverTimeInternal()
{
    int currTimeIter = 0;

    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Iterate over time..." << std::endl;
    }
    this->mParams.decreaseLevel();

/*------------------------------------------------------------------------*/
#ifdef VTRACE
    unsigned int mid;
    mid = VT_MARKER_DEF("Prblm_iterate", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Prblm_iterate_start");
#endif

    this->mTf = static_cast<CT>(0);
    for (std::size_t ii = 0; ii < this->mMemoryVectGradTf.size(); ++ii)
    {
        this->mMemoryVectGradTf[ii] = static_cast<CT>(0);
    }
    for (currTimeIter = 0; currTimeIter <= this->nTimesteps(); ++currTimeIter)
    {

        /*--------------------------------------------------------------------*/
        if ((this->params().rankOutput() == this->myRank()) &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * Process TIME ITERATE=" << currTimeIter << "..."
                      << std::endl;
        }
        this->mParams.decreaseLevel();

        /*--------------------------------------------------------------------*/
        this->evalKnownDfs(currTimeIter);
        if (0 < currTimeIter)
        {
            if (this->params().enabledAdolc())
            {
                this->traceUpdatePhase(currTimeIter);
            }
            this->evalUpdatePhase(currTimeIter);
            if (this->useLeapfrogIntegration())
            {
                // Swap old / new data fields, BUT DO NOT SWAP times.
                // ==> Swap times 2x.
                this->swapPointersOfUnknownDfs(); // 1st swap of times.
                this->swapTimesOfUnknownDfs();    // 2nd swap of times.
                if (this->params().enabledAdolc())
                {
                    this->traceUpdate2Phase(currTimeIter);
                }
                this->evalUpdate2Phase(currTimeIter);
            }
            // Compute target functional at each time step.
            if (this->params().isTfTimeDependent())
            {
                // Create a new ADOL-C tape in each time step.
                if (this->params().enabledAdolc())
                {
                    this->mIsTracedTfPhase = false;
                    this->traceTfPhase(currTimeIter);
                }
                this->evalTfPhase(currTimeIter);
            }
            else
            {
                if (currTimeIter == this->params().nTimesteps())
                {
                    if (this->params().enabledAdolc())
                    {
                        this->traceTfPhase(currTimeIter);
                    }
                    this->evalTfPhase(currTimeIter);
                }
            }
        }
        else
        {
            this->resetValsAtNodesOfUnknownDfs(currTimeIter);
            this->resetTimesOfUnknownDfs();
            if (this->params().enabledAdolc())
            {
                this->traceInitPbPhase(currTimeIter);
            }
            this->evalInitPbPhase();
        }
        this->writeAllDfs(currTimeIter);
        this->compErrOfUnknownDfs();
        this->swapPointersOfUnknownDfs();
        // Update times of old / new data fields AFTER swapping.
        this->updateTimesOfUnknownDfs();

        /*--------------------------------------------------------------------*/
        this->mParams.increaseLevel();
        if ((this->params().rankOutput() == this->myRank()) &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "   Processed." << std::endl;
        }
    }
#ifdef VTRACE
    mid = VT_MARKER_DEF("EndLoop", VT_MARKER_TYPE_HINT);
    VT_MARKER(mid, "Prblm_iterate_end");
#endif

    /*------------------------------------------------------------------------*/
    // Print some statistics.
    double tSum = 0.0;
    for (auto it = this->mWallClockTimes.begin();
         it != this->mWallClockTimes.end(); ++it)
    {
        tSum += it->second;
    }
    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << std::scientific << std::fixed
                  << " * Summary of Wall clock times(compNewTfAndGradTf):"
                  << std::endl;
        for (auto it = this->mWallClockTimes.begin();
             it != this->mWallClockTimes.end(); ++it)
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "       t(" << std::setw(11) << it->first
                      << ") = " << std::setw(8) << std::setprecision(5)
                      << it->second << " s (" << std::setw(5)
                      << std::setprecision(2) << it->second / tSum * 100.0
                      << " %%)" << std::endl;
        }
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "       --------------------------------------"
                  << std::endl
                  << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "       t(compTfGrdTf) = " << std::setw(8)
                  << std::setprecision(5) << tSum << " s."
                  << resetiosflags(std::ios::fixed | std::ios::showpoint)
                  << std::endl;
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Iterated." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void
Problem<OWNPRBLM, CT, DIM>::minimizeTargetfct(std::string const& optimizer)
{
    if (0 == optimizer.compare("lbfgsb"))
    {
        int optimVersion = 2; // Main version of optimizer L-BFGS-B.
        minimizeTargetfctUsingLbfgsb(optimVersion);
    }
    else
    {
        if (this->params().rankOutput() == this->myRank() &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * Optimizer " << optimizer << " not implemented."
                      << " Return." << std::endl;
        }
        return;
    }
}

/*******************************************************************************
 * GETTER METHODS.
 ******************************************************************************/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline bool const& Problem<OWNPRBLM, CT, DIM>::isTracedTfPhase() const
{
    return this->mIsTracedTfPhase;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const&
Problem<OWNPRBLM, CT, DIM>::tapeIdComputeTfPartitionInner() const
{
    return this->mTapeIdComputeTfPartitionInner;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const&
Problem<OWNPRBLM, CT, DIM>::tapeIdComputeTfPartitionBorder() const
{
    return this->mTapeIdComputeTfPartitionBorder;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline short int const&
Problem<OWNPRBLM, CT, DIM>::tapeIdComputeTfPartitionAll() const
{
    return this->mTapeIdComputeTfPartitionAll;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline int const& Problem<OWNPRBLM, CT, DIM>::nParamsToOptimize() const
{
    return this->mNparamsToOptimize;
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline std::vector<bool> const& Problem<OWNPRBLM, CT, DIM>::optimizeDf() const
{
    return this->mOptimizeDf;
}

/*******************************************************************************
 * WORK METHODS.
 ******************************************************************************/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::minimizeTargetfctUsingLbfgsb(
    int const& optimVersion)
{
    if (this->params().rankOutput() == this->myRank() &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Minimize target functional using L-BFGS-B..."
                  << std::endl;
    }
    this->mParams.decreaseLevel();

    /*------------------------------------------------------------------------*/
    // Optimizer (MPI LOCAL): Declarations and initializations.
    ScaFES::Timer timerMinimize;
    double tMinimize = 0.0;
    ScaFES::Timer timerCommunicate;
    ScaFES::Timer timerLbfgsb;
    timerMinimize.restart();
    int rankOptim = 0;

    /*------------------------------------------------------------------------*/
    // nParamsToOptimizeGlobal = Number of geometrical parameters. e.g.
    //
    // It is assumed that all data fields have the same number of nodes.
    // As all data fields are scalar and all (global) data fields live on
    // the global grid, this assumption is true in the moment.
    // ==> nParamsToOptimizeGlobal = Global number of grid nodes multiplied
    // by the number of those data fields which should be opimized.
    int nParamsToOptimizeGlobal = 0;
    if (this->isParamsBased())
    {
        nParamsToOptimizeGlobal = this->vectParamsCurr().size();
    }
    else
    {
        int kk = 0;
        int tmpNvars = 0;
        for (std::size_t jj = 0; jj < this->isKnownDf().size(); ++jj)
        {
            if (!this->isKnownDf().at(jj))
            {
                tmpNvars = this->vectUnknownDfsDomNew().at(kk).gridGlobal()
                               .discreteDomain().nNodes()
                               .size();
                ++kk;
            }
            if (this->optimizeDf().at(jj))
            {
                nParamsToOptimizeGlobal += tmpNvars;
            }
        }
    }

    /*------------------------------------------------------------------------*/
    // L-BFGS-B (MPI LOCAL): Declare quantities of L-BFGS-B
    // Declare variables for ALL processes such that LBFGS-B related variables
    // are known to all processes.
    // This must be done for ALL processes since there are several
    // "if (rank == rankOptim)" sections.
    // ==> Variables cannot be declared in the first local if section.
    // Instead, LBFGS-B related variables must be known to all processes.
    LbfgsbParams ppp;

    // Instead, LBFGS-B related variables must be known to all processes.
    // Initialize quantities of L-BFGS-B depending on the number of unknowns.
    if (rankOptim == this->myRank())
    {
        const int N_CORRECTIONS_IN_MATRIX = 5;
        ppp.initLbfgsbParams(nParamsToOptimizeGlobal, N_CORRECTIONS_IN_MATRIX,
                             optimVersion);
    }
    /*------------------------------------------------------------------------*/
    // (MPI LOCAL): Iterate.
    this->compNewTfAndGradTf();
    this->worldBarrier();

    /*------------------------------------------------------------------------*/
    // MPI: Prepare collecting and distributing.
    // Values: nDfsToOptimize approx 1 - 20
    //         nProcesses approx 100 - 10.000
    int nDfsToOptimize = this->getNdataFieldsForOptimization();
    std::vector<int> nParamsToOptimizeLocalOfDf;
    int** displs = 0;
    int** rcounts = 0;
    if (!(this->isParamsBased()))
    {
        this->compVectOfNelemsToOptimize(nParamsToOptimizeLocalOfDf,
                                         nDfsToOptimize);
        this->params().myWorld().compMatRcounts(rcounts,
                             nParamsToOptimizeLocalOfDf,
                             nDfsToOptimize);
        this->params().myWorld().compMatDisplacements(displs,
                                                      rcounts, nDfsToOptimize);
    }

    /*------------------------------------------------------------------------*/
    if (this->params().rankOutput() == this->myRank() &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * (Local) Value of target functional: " << this->mTf
                  << std::endl;
    }

    /*------------------------------------------------------------------------*/
    // Collect local values of target functional to global value.
    timerCommunicate.restart();
    this->params().myWorld().collectScalar(ppp.scalarTf,
                                           static_cast<double>(this->mTf),
                                           rankOptim);

    /*------------------------------------------------------------------------*/
    // Collect local arrays in corresponding global ones.
    if (this->isParamsBased())
    {
        // REMARK: Initial guess is already the same on all processes.
        // => No mpi_reduce() necessary.
        if (rankOptim == this->myRank())
        {
            for (int ii = 0; ii < this->nColumns(); ++ii)
            {
                ppp.memoryIterate[ii] = this->mMemoryVectIterate[ii];
            }
        }
        if (rankOptim == this->myRank())
        {
            for (int ii = 0; ii < this->nColumns(); ++ii)
            {
                ppp.memoryGradTf[ii] = 0.0;
            }
        }
        this->params().myWorld().collectVector(
                this->mMemoryVectGradTf,
                ppp.memoryGradTf,
                this->nColumns(),
                rankOptim);
    }
    else
    {
        this->params().myWorld().assembleSubvectors(ppp.memoryIterate,
                                                this->mMemoryVectIterate,
                             rcounts, displs, nParamsToOptimizeLocalOfDf,
                             nParamsToOptimize(), nDfsToOptimize, rankOptim);
        this->params().myWorld().assembleSubvectors(ppp.memoryGradTf,
                                                this->mMemoryVectIterateGradTf,
                             rcounts, displs, nParamsToOptimizeLocalOfDf,
                             nParamsToOptimize(), nDfsToOptimize, rankOptim);
    }
    this->mWallClockTimesOptim["communic"] += timerCommunicate.elapsed();

    /*------------------------------------------------------------------------*/
    // Optimizer (MPI LOCAL): Loop.
    bool iterateOptimLoop = true;
    int currIteration = 0;
    do
    {

        if (this->params().rankOutput() == this->myRank() &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * Process OPTIM_ITERATION=" << currIteration
                      << std::endl;
        }
        this->mParams.decreaseLevel();

        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank() && (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * (Global) parameter vector:" << std::endl;
            for (int ii = 0; ii < this->nColumns(); ++ii)
            {
                std::cout << " p[ " << ii
                          << " ] = " << ppp.memoryIterate.at(ii) << std::endl;
            }
        }
        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank() && (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * (Global) Value of target functional: "
                      << ppp.scalarTf<< std::endl;
        }

        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank() && (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * (Global) Value of gradient of tf: " << std::endl;
            for (int ii = 0; ii < this->nColumns(); ++ii)
            {
                std::cout << " gradTf[ " << ii
                          << " ] = " << ppp.memoryGradTf.at(ii) << std::endl;
            }
        }

        std::string strtask("");
        strtask.resize(60);
#ifdef SCADOPT_HAVE_LBFGSB
        /*--------------------------------------------------------------------*/
        // L-BFGS-B (MPI GLOBAL): Call FORTRAN routine.
        // void setulb_(int*, int*, double*, double*, double*, int*, double*,
        //             double*, double*, double*, double*, int*, char*, int*,
        //             char*, bool*, int*, double*)
        if (this->params().rankOutput() == this->myRank() &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << " * Call L-BFGS-B routine setulb()..." << std::endl;
        }
        this->mParams.decreaseLevel();
        timerLbfgsb.restart();

        /*--------------------------------------------------------------------*/
        if (rankOptim == this->myRank())
        {
            double* ptr2memoryIterateGlobal = ppp.memoryIterate.data();
            double* ptr2lowerBound = ppp.lowerBound.data();
            double* ptr2upperBound = ppp.upperBound.data();
            int* ptr2typeOfBounds = ppp.typeOfBounds.data();
            double* ptr2memoryGradTfGlobal = ppp.memoryGradTf.data();
            double* ptr2wa = ppp.wa.data();
            int* ptr2iwa = ppp.iwa.data();
            char* ptr2csave = ppp.csave.data();
            int* ptr2lsave = ppp.lsave.data();
            int* ptr2isave = ppp.isave.data();
            double* ptr2dsave = ppp.dsave.data();

            setulb_(&nParamsToOptimizeGlobal, &(ppp.nCorrectionsInMatrix),
                    ptr2memoryIterateGlobal, ptr2lowerBound, ptr2upperBound,
                    ptr2typeOfBounds, &(ppp.scalarTf), ptr2memoryGradTfGlobal,
                    &(ppp.factr), &(ppp.pgtol),
                    ptr2wa, ptr2iwa, ppp.task, &(ppp.info),
                    ptr2csave,
                    ptr2lsave, ptr2isave, ptr2dsave);
            for (int ii = 0; ii < 60; ++ii)
            {
                strtask[ii] = ppp.task[ii];
            }
        }

        // Broadcast task to all other processes.
        this->params().myWorld().barrier();
        ScaFES::broadcastChars(this->params().myWorld(), ppp.task, 60+1,
                               rankOptim);
        // Alternative:
        // ScaFES::broadcast(this->params().myWorld(), strtask, rankOptim);

       // Set task variable to string variant strtask on all processes.
        strtask = ppp.task;

        this->mWallClockTimesOptim["lbfgsb"] += timerLbfgsb.elapsed();
        this->mParams.increaseLevel();
        if (this->params().rankOutput() == this->myRank() &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "     * New task: " << strtask << std::endl
                      << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "   Called." << std::endl;
        }
#endif

        /*--------------------------------------------------------------------*/
        // L-BFGS-B (MPI LOCAL):
        // Case: Function and its gradient must be calculated
        // at current iterate.
        // if (0 == strncmp(task, "FG", 2)) {
        if (0 == strncmp(strtask.c_str(), "FG", 2))
        {
            // Distribute global iterate to local arrays.
            timerCommunicate.restart();
            if (this->isParamsBased())
            {
                if (rankOptim == this->myRank())
                {
                    // Copy the parameter based vector got from the setulb().
                    for (int ll = 0; ll < this->nParamsToOptimize(); ++ll)
                    {
                        this->mMemoryVectIterate[ll] = ppp.memoryIterate[ll];
                    }
                    for (int ll = 0; ll < this->nParamsToOptimize(); ++ll)
                    {
                        this->mVectParamsCurr[ll] =
                            this->mMemoryVectIterate[ll];
                    }
                    // The vector of the geometrical parameters usually
                    // consists only of a small number (1-50) of elements.
                    // --> Just broadcast the vector to all MPI processes.
                    // mVectParamsCurr is used instead of mMemoryVectIterate
                    // because boost::mpi can handle variables
                    // of type std::vector.
                    ScaFES::broadcastVector(this->params().myWorld(),
                                            this->mVectParamsCurr, rankOptim);
                }
            }
            else
            {
                this->params().myWorld().partitionGlobalvector(
                    this->mMemoryVectIterate, ppp.memoryIterate, rcounts,
                    displs, nParamsToOptimizeLocalOfDf,
                    this->nParamsToOptimize(), nDfsToOptimize, rankOptim);
                // mVectIterate works on nodes of grid memNormal!
                int idDf = 0;
                int idOptimizeAndUnknownDf = 0;
                for (std::size_t ii = 0; ii < this->optimizeDf().size(); ++ii)
                {
                    if (!this->isKnownDf().at(ii))
                    {
                        if (this->optimizeDf().at(ii))
                        {
                            this->mVectUnknownDfsDomOld
                                [idDf].setValuesAtMemNormal(
                                    this->mVectIterate[idOptimizeAndUnknownDf]);
                            ++idOptimizeAndUnknownDf;
                        }
                        ++idDf;
                    }
                }
            }
            this->mWallClockTimesOptim["communic"] +=
                timerCommunicate.elapsed();

            /*----------------------------------------------------------------*/
            // mVectUnknownDfsDomOld has to contain the values of the
            // current iterate.
            this->compNewTfAndGradTf();

            /*----------------------------------------------------------------*/
            // Collect local scalar and arrays in corresponding global ones.
            timerCommunicate.restart();
            this->params().myWorld().collectScalar(ppp.scalarTf,
                                                   this->mTf, rankOptim);
            if (this->isParamsBased())
            {
                // REMARK: Initial guess is already the same on all processes.
                // => No mpi_reduce() necessary.
                if (rankOptim == this->myRank())
                {
                    for (int ii = 0; ii < this->nColumns(); ++ii)
                    {
                        ppp.memoryIterate[ii] = this->mMemoryVectIterate[ii];
                    }
                }
                this->params().myWorld().collectVector(
                    this->mMemoryVectGradTf,
                    ppp.memoryGradTf,
                    this->nColumns(),
                    rankOptim);
            }
            else
            {
                this->params().myWorld().assembleSubvectors(
                    ppp.memoryIterate, this->mMemoryVectIterate, rcounts,
                    displs, nParamsToOptimizeLocalOfDf,
                    this->nParamsToOptimize(), nDfsToOptimize, rankOptim);
                this->params().myWorld().assembleSubvectors(
                    ppp.memoryGradTf, this->mMemoryVectIterateGradTf, rcounts,
                    displs, nParamsToOptimizeLocalOfDf,
                    this->nParamsToOptimize(), nDfsToOptimize, rankOptim);
            }
            this->mWallClockTimesOptim["communic"] +=
                timerCommunicate.elapsed();
        }
        /*--------------------------------------------------------------------*/
        // Case: New iterate was computed, the user can decide:
        // Stop or continue the iteration.
        else if (0 == strncmp(strtask.c_str(), "NEW_X", 5))
        {
            // Currently, the iteration will be continued.
        }
        /*--------------------------------------------------------------------*/
        // Case: Iteration will be terminated.
        else
        {
            iterateOptimLoop = false;
        }

        this->mParams.increaseLevel();
        if (this->params().rankOutput() == this->myRank() &&
            (0 < this->params().indentDepth()))
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "   Processed." << std::endl;
        }
        ++currIteration;
    } while (iterateOptimLoop
             && currIteration < this->params().nOptimIterationsMax());

    /*------------------------------------------------------------------------*/
    // Print some statistics.
    double tSum = 0.0;
    for (auto it = this->mWallClockTimes.begin();
         it != this->mWallClockTimes.end(); ++it)
    {
        tSum += it->second;
    }
    for (auto it = this->mWallClockTimesOptim.begin();
         it != this->mWallClockTimesOptim.end(); ++it)
    {
        tSum += it->second;
    }
    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                        this->myRank(), "SCADOPT")
                  << std::scientific << std::fixed
                  << " * Summary of Wall clock times(optimize):" << std::endl;
        for (auto it = this->mWallClockTimes.begin();
             it != this->mWallClockTimes.end(); ++it)
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "       t(" << std::setw(11) << it->first
                      << ") = " << std::setw(8) << std::setprecision(5)
                      << it->second << " s (" << std::setw(5)
                      << std::setprecision(2) << it->second / tSum * 100.0
                      << " %%)" << std::endl;
        }
        for (auto it = this->mWallClockTimesOptim.begin();
             it != this->mWallClockTimesOptim.end(); ++it)
        {
            std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                           this->myRank(), "SCADOPT")
                      << "       t(" << std::setw(11) << it->first
                      << ") = " << std::setw(8) << std::setprecision(5)
                      << it->second << " s (" << std::setw(5)
                      << std::setprecision(2) << it->second / tSum * 100.0
                      << " %%)" << std::endl;
        }
        tMinimize = timerMinimize.elapsed();
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "       t(   sumItems) = " << std::setw(8)
                  << std::setprecision(5) << tSum << " s."
                  << resetiosflags(std::ios::fixed | std::ios::showpoint)
                  << std::endl
                  << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " ----> t(   optimize) = " << std::setw(8)
                  << std::setprecision(5) << tMinimize << " s. <-------"
                  << resetiosflags(std::ios::fixed | std::ios::showpoint)
                  << std::endl;
    }
    this->mParams.increaseLevel();
    if ((this->params().rankOutput() == this->myRank()) &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Minimized." << std::endl;
    }
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::compVectOfNelemsToOptimize(
    std::vector<int>& nParamsToOptimizeLocalOfDf,
    int const& nDfsToOptimize
)
{
    nParamsToOptimizeLocalOfDf.resize(nDfsToOptimize, 0);
//     for (int jj = 0; jj < nDfsToOptimize; ++jj)
//     {
//         nParamsToOptimizeLocalOfDf[jj] = 0;
//     }
    int idDf = 0;
    int idOptimizeAndUnknownDf = 0;
    int tmpNvars = 0;
    for (std::size_t jj = 0; jj < this->isKnownDf().size(); ++jj)
    {
        if (!this->isKnownDf().at(jj))
        {
            tmpNvars = this->vectUnknownDfsDomNew().at(idDf).memNormal()
                .nNodesSub().size();
            if (this->optimizeDf().at(jj))
            {
                nParamsToOptimizeLocalOfDf[idOptimizeAndUnknownDf] = tmpNvars;
                ++idOptimizeAndUnknownDf;
            }
            ++idDf;
        }
    }
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline int Problem<OWNPRBLM, CT, DIM>::getNdataFieldsForOptimization() const
{
    int nDataFieldsForOptimization = 0;
    for (std::size_t ii = 0; ii < this->optimizeDf().size(); ++ii)
    {
        if (this->optimizeDf().at(ii))
        {
            ++nDataFieldsForOptimization;
        }
    }

    return nDataFieldsForOptimization;
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void
Problem<OWNPRBLM, CT, DIM>::evalInner(SCADOPT_VectDf<TT>& dfDep,
                                      SCADOPT_IntNtuple const& idxNode,
                                      int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template evalInner<TT>(dfDep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void
Problem<OWNPRBLM, CT, DIM>::evalBorder(SCADOPT_VectDf<TT>& dfDep,
                                       SCADOPT_IntNtuple const& idxNode,
                                       int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template evalBorder<TT>(dfDep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initInner(
    SCADOPT_VectDf<TT>& dfDep, std::vector<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template initInner<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initBorder(
    SCADOPT_VectDf<TT>& dfDep, std::vector<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template initBorder<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initInner(
    SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template initInner<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::initBorder(
    SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template initBorder<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateInner(
    SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template updateInner<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateBorder(
    SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template updateBorder<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateInner2(
    SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template updateInner2<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::updateBorder2(
    SCADOPT_VectDf<TT>& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template updateBorder2<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::compTfInner(
    TT& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template compTfInner<TT>(dfDep, dfIndep, idxNode, timeIter);
}
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
template <typename TT>
inline void Problem<OWNPRBLM, CT, DIM>::compTfBorder(
    TT& dfDep, SCADOPT_VectDf<TT> const& dfIndep,
    SCADOPT_IntNtuple const& idxNode, int const& timeIter)
{
    static_cast<OWNPRBLM*>(this)
        ->template compTfBorder<TT>(dfDep, dfIndep, idxNode, timeIter);
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::evalTfPhase(int const& timeIter)
{
    if (this->params().rankOutput() == this->myRank() &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << " * Evaluate target functional..." << std::endl;
    }
    this->mParams.decreaseLevel();

    CT tfTmp = static_cast<CT>(0);
    std::vector<CT> tfGradTmp(this->mMemoryVectGradTf.size(),
                              static_cast<CT>(0));

    if (this->useAsynchronMode())
    {
        this->evalTypeThreeAsynchronously(
            "compTf", tfTmp, tfGradTmp, this->vectUnknownDfsDomNew(),
            this->mVectGradUnknownDfsDomNew, timeIter,
            this->tapeIdComputeTfPartitionInner(),
            this->tapeIdComputeTfPartitionBorder(),
            &Problem<OWNPRBLM, CT, DIM>::template compTfInner<CT>,
            &Problem<OWNPRBLM, CT, DIM>::template compTfBorder<CT>);
    }
    else
    {
        this->evalTypeThreeSynchronously(
            "compTf", tfTmp, tfGradTmp, this->vectUnknownDfsDomNew(),
            this->mVectGradUnknownDfsDomNew, this->vectUnknownDfsBdryNew(),
            this->mVectGradUnknownDfsBdryNew, timeIter,
            this->tapeIdComputeTfPartitionAll(),
            &Problem<OWNPRBLM, CT, DIM>::template compTfInner<CT>,
            &Problem<OWNPRBLM, CT, DIM>::template compTfBorder<CT>);
    }

    if (this->params().isTfTimeDependent())
    {
        this->mTf += tfTmp;
        for (std::size_t ii = 0; ii < tfGradTmp.size(); ++ii)
        {
            this->mMemoryVectGradTf[ii] += tfGradTmp[ii];
        }
    }
    else
    {
        this->mTf = tfTmp;
        for (std::size_t ii = 0; ii < tfGradTmp.size(); ++ii)
        {
            this->mMemoryVectGradTf[ii] = tfGradTmp[ii];
        }
    }

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank() &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Computed." << std::endl;
    }
}

/*----------------------------------------------------------------------------*/
/*----------------------------------------------------------------------------*/
template <class OWNPRBLM, typename CT, std::size_t DIM>
inline void Problem<OWNPRBLM, CT, DIM>::traceTfPhase(int const& timeIter)
{
    if (this->isTracedTfPhase())
    {
        return;
    }

    std::string whichPhase = "compTf";

    if (this->params().rankOutput() == this->myRank() &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT") << " * Trace "
                  << whichPhase << " phase..." << std::endl;
    }
    this->mParams.decreaseLevel();

#ifdef SCADOPT_HAVE_ADOLC
    CT tfTmp = static_cast<CT>(0);

    this->template traceTypeThree<adouble>(
        whichPhase, tfTmp, this->vectUnknownDfsDomNew(),
        this->vectUnknownDfsBdryNew(), timeIter,
        this->tapeIdComputeTfPartitionInner(),
        this->tapeIdComputeTfPartitionBorder(),
        this->tapeIdComputeTfPartitionAll(),
        &Problem<OWNPRBLM, CT, DIM>::template compTfInner<adouble>,
        &Problem<OWNPRBLM, CT, DIM>::template compTfBorder<adouble>);
    // //     if (this->params().isTfTimeDependent()) {
    // //         this->mTf += tfTmp;
    // //     }
    // //     else {
    // //         this->mTf = tfTmp;
    // //     }
    this->mIsTracedTfPhase = true;
#else
    std::ignore = timeIter;
#endif

    this->mParams.increaseLevel();
    if (this->params().rankOutput() == this->myRank() &&
        (0 < this->params().indentDepth()))
    {
        std::cout << SCADOPT::getPrefix(this->params().indentDepth(),
                                       this->myRank(), "SCADOPT")
                  << "   Traced." << std::endl;
    }
}
} // End of namespace.
#endif


