#!/bin/bash
#
# Call: ./scadoptcc.sh <source.cpp>
# Parameters: source.cpp: Main source file containing SCADOPT related objects.
#
# Shell script for compiling and linking SCADOPT applications.
# The shell script can be configured via environment variables prefixed with
# SCADOPTCC. All environment variables are listed in the following
# and have to be exported.
#
# * C++11 compiler. <g++, icpc, pgc++...> default=g++
#   SCADOPTCC_CXX='g++';
#
# * Compiler flags. default="-DNDEBUG -O3 -Wall"
#   SCADOPTCC_CXXFLAGS="-DNDEBUG -O3 -Wall";
#
# * Enable MPI? <no/yes>. default=yes
#   SCADOPTCC_ENABLE_MPI="yes";
#
# * Only needed if SCADOPTCC_ENABLE_MPI="yes": default=/usr
#   Root directory of MPI installation. <path>
#   SCADOPTCC_MPI_ROOT="/usr";
#
# * Root directory of Boost installation. <path> default=/usr
#   SCADOPTCC_BOOST_ROOT="/usr";
#
# * Library directory of Boost installation default=/usr/lib64.
#   SCADOPTCC_BOOST_LIB="/usr/lib64"; <path>
#
# * Enable Boost.MPI? <no/yes>. default=yes
#   SCADOPTCC_ENABLE_BOOST_MPI="yes";
#
# * Enable OpenMP? <no/yes>. default=no
#   SCADOPTCC_ENABLE_OPENMP="no";
#
# * Enable NetCDF? <no/yes>. default=no
#   SCADOPTCC_ENABLE_NETCDF="no";
#
# * Enable LBFGSB? <no/yes>. default=yes
#   SCADOPTCC_ENABLE_LBFGSB="yes";
#
# * Only needed if SCADOPTCC_ENABLE_NETCDF="yes":
#   Root directory of NetCDF installation. <path>
#   SCADOPTCC_NETCDF_ROOT="/usr";
#
# * Only needed if SCADOPTCC_ENABLE_NETCDF="yes":
#   Root directory of Hdf5 installation. <path> default=/usr
#   SCADOPTCC_HDF5_ROOT="/usr";
#
# * Enable ADOL-C? <no/yes>. default=no
#   SCADOPTCC_ENABLE_ADOLC="no";
#
# * Only needed if SCADOPTCC_ENABLE_ADOLC="yes": default=/usr
#   Root directory of ADOL-C installation. <path>
#   SCADOPTCC_ADOLC_ROOT="/usr";
#
# * Only needed if SCADOPTCC_ENABLE_LBFGSB="yes": default=/usr
#   Root directory of LBFGS-B installation. <path>
#   SCADOPTCC_LBFGSB_ROOT="/usr";
#
# * NECESSARY: Root directory of SCAFES installation. <path> default=/usr
#   SCADOPTCC_SCAFES_ROOT="/usr";
#
# * NECESSARY: Root directory of SCADOPT installation. <path> default=/usr
#   SCADOPTCC_SCADOPT_ROOT="/usr";
#
# SCADOPT
# Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
# For details, see the files COPYING and LICENSE in the base directory
# of the package.

(
#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_CXX" == "" ] ; then
    SCADOPTCC_CXX="g++";
    echo "* WARNING: Use default value SCADOPTCC_CXX=$SCADOPTCC_CXX."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_CXXFLAGS" == "" ] ; then
    SCADOPTCC_CXXFLAGS="-DNDEBUG -O3 -Wall";
    echo "* WARNING: Use default value SCADOPTCC_CXXFLAGS=$SCADOPTCC_CXXFLAGS."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ENABLE_MPI" == "" ] ; then
    SCADOPTCC_ENABLE_MPI="yes";
    echo "* WARNING: Use default value SCADOPTCC_ENABLE_MPI=$SCADOPTCC_ENABLE_MPI."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ENABLE_BOOST_MPI" == "" ] ; then
    SCADOPTCC_ENABLE_BOOST_MPI="yes";
    echo "* WARNING: Use default value SCADOPTCC_ENABLE_BOOST_MPI=$SCADOPTCC_ENABLE_BOOST_MPI."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ENABLE_ADOLC" == "" ] ; then
    SCADOPTCC_ENABLE_ADOLC="no";
    echo "* WARNING: Use default value SCADOPTCC_ENABLE_ADOLC=$SCADOPTCC_ENABLE_ADOLC."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ENABLE_OPENMP" == "" ] ; then
    SCADOPTCC_ENABLE_OPENMP="no";
    echo "* WARNING: Use default value SCADOPTCC_ENABLE_OPENMP=$SCADOPTCC_ENABLE_OPENMP."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ENABLE_NETCDF" == "" ] ; then
    SCADOPTCC_ENABLE_NETCDF="no";
    echo "* WARNING: Use default value SCADOPTCC_ENABLE_NETCDF=$SCADOPTCC_ENABLE_NETCDF."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ENABLE_LBFGSB" == "" ] ; then
    SCADOPTCC_ENABLE_LBFGSB="yes";
    echo "* WARNING: Use default value SCADOPTCC_ENABLE_LBFGSB=$SCADOPTCC_ENABLE_LBFGSB."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_BOOST_ROOT" == "" ] ; then
    SCADOPTCC_BOOST_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_BOOST_ROOT=$SCADOPTCC_BOOST_ROOT."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_BOOST_LIB" == "" ] ; then
    SCADOPTCC_BOOST_LIB="/usr/lib64";
    echo "* WARNING: Use default value SCADOPTCC_BOOST_LIB=$SCADOPTCC_BOOST_LIB."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_ADOLC_ROOT" == "" ] ; then
    SCADOPTCC_ADOLC_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_ADOLC_ROOT=$SCADOPTCC_ADOLC_ROOT."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_NETCDF_ROOT" == "" ] ; then
    SCADOPTCC_NETCDF_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_NETCDF_ROOT=$SCADOPTCC_NETCDF_ROOT."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_HDF5_ROOT" == "" ] ; then
    SCADOPTCC_HDF5_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_HDF5_ROOT=$SCADOPTCC_HDF5_ROOT."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_LBFGSB_ROOT" == "" ] ; then
    SCADOPTCC_LBFGSB_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_LBFGSB_ROOT=$SCADOPTCC_LBFGSB_ROOT."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_SCAFES_ROOT" == "" ] ; then
    SCADOPTCC_SCAFES_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_SCAFES_ROOT=$SCADOPTCC_SCAFES_ROOT."
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTCC_SCADOPT_ROOT" == "" ] ; then
    SCADOPTCC_SCADOPT_ROOT="/usr";
    echo "* WARNING: Use default value SCADOPTCC_SCADOPT_ROOT=$SCADOPTCC_SCADOPT_ROOT."
fi

# From here on, all variables should be defined.
set -u

#------------------------------------------------------------------------------#
declare local SCAFES_INC="$SCADOPTCC_SCAFES_ROOT/include"
declare local SCAFES_LIB="$SCADOPTCC_SCAFES_ROOT/lib"
declare local SCADOPT_INC="$SCADOPTCC_SCADOPT_ROOT/include"
declare local SCADOPT_LIB="$SCADOPTCC_SCADOPT_ROOT/lib"
declare local BOOST_INC="$SCADOPTCC_BOOST_ROOT/include"
declare local BOOST_LIB="$SCADOPTCC_BOOST_LIB"
declare local ADOLC_INC="$SCADOPTCC_ADOLC_ROOT/include"
declare local ADOLC_LIB="$SCADOPTCC_ADOLC_ROOT/lib"
declare local NETCDF_INC="$SCADOPTCC_NETCDF_ROOT/include"
declare local NETCDF_LIB="$SCADOPTCC_NETCDF_ROOT/lib"

#------------------------------------------------------------------------------#
declare local SCAFES_INC_DIR="-I${SCAFES_INC}"
declare local SCAFES_LDFLAGS="-L${SCAFES_LIB}"
declare local SCAFES_LIBADD="-lscafes"
declare local SCADOPT_INC_DIR="-I${SCADOPT_INC}"
declare local SCADOPT_LDFLAGS="-L${SCADOPT_LIB}"
declare local SCADOPT_LIBADD="-lscadopt"
declare local BOOST_INC_DIR="-I${BOOST_INC}"
declare local BOOST_LDFLAGS="-L${BOOST_LIB}"
declare local BOOST_LIBADD="-lboost_serialization"
BOOST_LIBADD="${BOOST_LIBADD} -lboost_regex"
BOOST_LIBADD="${BOOST_LIBADD} -lboost_program_options"

if [ ${SCADOPTCC_ENABLE_BOOST_MPI} == "yes" ] ; then
    BOOST_LIBADD="${BOOST_LIBADD} -lboost_mpi"
fi

declare local ADOLC_INC_DIR=""
declare local ADOLC_LDFLAGS=""
declare local ADOLC_LIBADD=""
if [ ${SCADOPTCC_ENABLE_ADOLC} == "yes" ] ; then
    ADOLC_INC_DIR="-I${ADOLC_INC}"
    ADOLC_LDFLAGS="-L${ADOLC_LIB}"
    ADOLC_LIBADD="-ladolc"
fi

declare local NETCDF_INC_DIR=""
declare local NETCDF_LDFLAGS=""
declare local NETCDF_LIBADD=""
if [ ${SCADOPTCC_ENABLE_NETCDF} == "yes" ] ; then
   NETCDF_INC_DIR="-I${NETCDF_INC}"
   NETCDF_LDFLAGS="-L${NETCDF_LIB}"
   NETCDF_LIBADD="-lnetcdf"
fi

declare local LBFGSB_INC_DIR=""
declare local LBFGSB_LDFLAGS=""
declare local LBFGSB_LIBADD=""
if [ ${SCADOPTCC_ENABLE_LBFGSB} == "yes" ] ; then
   LBFGSB_INC_DIR="-I${LBFGSB_INC}"
   LBFGSB_LDFLAGS="-L${LBFGSB_LIB}"
   LBFGSB_LIBADD="-llbfgsb"
fi

declare local OPENMP_CXXFLAGS=""
if [ ${SCADOPTCC_ENABLE_OPENMP} == "yes" ] ; then
    OPENMP_CXXFLAGS="-fopenmp"
fi

#------------------------------------------------------------------------------#
declare local CPPFLAGS="${SCAFES_INC_DIR}"
CPPFLAGS="${CPPFLAGS} ${BOOST_INC_DIR}"
CPPFLAGS="${CPPFLAGS} ${ADOLC_INC_DIR}"
CPPFLAGS="${CPPFLAGS} ${LBFGSB_INC_DIR}"
CPPFLAGS="${CPPFLAGS} ${NETCDF_INC_DIR}"
CPPFLAGS="${CPPFLAGS} ${OPENMP_CXXFLAGS}"
CPPFLAGS="${CPPFLAGS} -std=c++11"
declare local LDFLAGS="${SCAFES_LDFLAGS} ${SCAFES_LIBADD}"
LDFLAGS="${LDFLAGS} ${SCADOPT_LDFLAGS} ${SCADOPT_LIBADD}"
LDFLAGS="${LDFLAGS} ${BOOST_LDFLAGS} ${BOOST_LIBADD}"
LDFLAGS="${LDFLAGS} ${ADOLC_LDFLAGS} ${ADOLC_LIBADD}"
LDFLAGS="${LDFLAGS} ${NETCDF_LDFLAGS} ${NETCDF_LIBADD}"
LDFLAGS="${LDFLAGS} ${OPENMP_CXXFLAGS}"

declare local MYCXX="${SCADOPTCC_CXX}"
if [ ${SCADOPTCC_ENABLE_MPI} == "yes" ] ; then
   MYCXX="mpic++";
fi
if [ ${SCADOPTCC_ENABLE_BOOST_MPI} == "yes" ] ; then
   MYCXX="mpic++";
fi

execString="${CPPFLAGS} ${SCADOPTCC_CXXFLAGS} ${LDFLAGS} $1";
echo "+++ ${MYCXX} ${execString} +++"
${MYCXX} ${execString}
)
