#!/bin/bash
#
# Call: ./scadoptrun.sh
# Shell script for running a SCADOPT application which was configured via
# environment variables prefixed with SCADOPTRUN.
#
# * NECESSARY: Name of SCADOPT executable. <string>
#   SCADOPTRUN_NAME_EXECUTABLE="$1";
#
# * NECESSARY: Space dimension of problem. <int>
#   SCADOPTRUN_SPACE_DIM="3";
#
# * Number of grid nodes in each direction. n1xn2x...xnd <string>
#   SCADOPTRUN_N_NODES="8x8x8";
#
# * Direction into which the grid can be partitioned when using the RCB
#   algorithm for the domain decomposition. d1xd2x...xdd <string>
#   SCADOPTRUN_PARTITION_GRID="0x0x1";
#
# * Coordinates of left lower point of the global computational domain.
#   s1xs2x...xsd with s_i \in R <string>
#   SCADOPTRUN_COORD_NODE_FIRST="0x0x0";
#
# * Coordinates of right upper point of the global computational domain.
#   e1xe2x...xed with e_i \in R <string>
#   SCADOPTRUN_COORD_NODE_LAST="1x1x1";
#
# * Number of timesteps. <int>
#   SCADOPTRUN_N_TIMESTEPS="1";
#
# * Number of snapshots when writing data files. <int>
#   SCADOPTRUN_N_SNAPSHOTS="1";
#
# * Start time t_S of time interval [t_S; t_E]. <double>
#   SCADOPTRUN_START_TIME="0";
#
# * End time t_E of time interval [t_S; t_E]. <double>
#   SCADOPTRUN_END_TIME="1";
#
# * Output level <int>
#   SCADOPTRUN_OUTPUT_LEVEL="5";
#
# * Number of layers at the global boundary <int>
#   SCADOPTRUN_NLAYERS_AT_BORDER="1";
#
# * Name of kind file if boundary description(=kind field) should be read in.
#   <string>
#   SCADOPTRUN_NAME_KINDFILE="";
#
# * Should the kind field be written to a file? <NO/YES>
#   SCADOPTRUN_WRITE_KINDFILE="NO";
#
#   Name of the configuration file (ini format) <string>
#   SCADOPTRUN_NAME_CONFIGFILE="";
#
# * Enable ADOL-C for evaluating function (and eventually computing
#   derivatives)? <NO/YES>
#   SCADOPTRUN_ENABLE_ADOLC="NO";
#
# * Enable asynchronous MPI communication for synchronisation of fields?
#   <NO/YES>
#   SCADOPTRUN_ASYNCHRON_MODE="YES";
#
# * Compute gradients using ADOL-C? <NO/YES>
#   SCADOPTRUN_COMP_GRADIENTS="NO"
#
# * Maximal number of iterations of the optimizer loop. <int>
#   SCADOPTRUN_N_OPTIM_ITERATIONS_MAX="50"
#
# * Name of directory in which results should be written <string>.
#   SCADOPTRUN_RESULTS_DIR="./";
#
# * Check results of SCADOPT application? <NO/YES>
#   SCADOPTRUN_CHECK_RESULTS="NO";
#
# * Run mode of scadoptrun.sh script?
#   DEBUG, DEBUG_MPI, NORMAL, LSF, SLURM <string>
#   SCADOPTRUN_RUN_MODE="DEBUG";
#
# * Number of runs with different number of MPI processes. <int>
#   SCADOPTRUN_N_TESTS_MPI="1";
#
# * Number of runs with different number of OpenMP threads. <int>
#   SCADOPTRUN_N_TESTS_OPENMP="1";
#
# * Number of MPI processes during the first run. <int>
#   SCADOPTRUN_N_PROCESSES_MPI_START="1";
#
# * Number of OpenMP threads during the first run. <int>
#   SCADOPTRUN_N_THREADS_OPENMP_START="1";
#
# * Type of scaling test <string> <STRONG/WEAK>
#   SCADOPTRUN_TYPE_SCALINGTEST="STRONG";
#
# * Name of machine on which application will be run <string>
#   SCADOPTRUN_MACHINE_NAME="LOCAL";
#
# * Number of nodes of the machine. <int>
#   SCADOPTRUN_MACHINE_N_NODES="1";
#
# * Number of cores per node of the machine. <int>
#   SCADOPTRUN_MACHINE_N_CORES_PER_NODE="1";
#
# * Name of the project of the machine. <string>
#   SCADOPTRUN_MACHINE_NAME_PROJECT="";
#
# * Name of reservation on the machine. <string>
#   SCADOPTRUN_MACHINE_NAME_RESERVATION="";
#
# * Desired memory per core at the machine in MB. <int>
#   SCADOPTRUN_MACHINE_MEMORY_PER_CORE="900M";
#
# SCADOPT
# Copyright (c) 2011-2014, ZIH, TU Dresden, Federal Republic of Germany.
# For details, see the files COPYING and LICENSE in the base directory
# of the package.

#ATCH --nodes=8
#ATCH --ntasks=16
#ATCH --time=00:10:00
#ATCH -p sandy
#ATCH -J heatEqn
#ATCH --mem-per-cpu=500M
#ATCH -A zihforschung

(

#------------------------------------------------------------------------------#
# Variables which must be set.
#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_NAME_EXECUTABLE" == "" ] ; then
    echo "* ERROR: SCADOPTRUN_NAME_EXECUTABLE was not set."
    exit 22;
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_SPACE_DIM" == "" ] ; then
    echo "* ERROR: SCADOPTRUN_SPACE_DIM was not set."
    exit 22;
fi

#------------------------------------------------------------------------------#
declare    local nameProblem=${SCADOPTRUN_NAME_EXECUTABLE};
declare -i local spaceDim=$SCADOPTRUN_SPACE_DIM;

#------------------------------------------------------------------------------#
# Variables which can be set.
#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_TESTS_MPI" == "" ] ; then
    SCADOPTRUN_N_TESTS_MPI="1";
    echo "* WARNING: Use default value SCADOPTRUN_N_TESTS_MPI=$SCADOPTRUN_N_TESTS_MPI"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_TESTS_OPENMP" == "" ] ; then
    SCADOPTRUN_N_TESTS_OPENMP="1";
    echo "* WARNING: Use default value SCADOPTRUN_N_TESTS_OPENMP=$SCADOPTRUN_N_TESTS_OPENMP"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_TYPE_SCALINGTEST" == "" ] ; then
    SCADOPTRUN_TYPE_SCALINGTEST="STRONG";
    echo "* WARNING: Use default value SCADOPTRUN_TYPE_SCALINGTEST=$SCADOPTRUN_TYPE_SCALINGTEST"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_PROCESSES_MPI_START" == "" ] ; then
    SCADOPTRUN_N_PROCESSES_MPI_START="1";
    echo "* WARNING: Use default value SCADOPTRUN_N_PROCESSES_MPI_START=$SCADOPTRUN_N_PROCESSES_MPI_START"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_THREADS_OPENMP_START" == "" ] ; then
    SCADOPTRUN_N_THREADS_OPENMP_START="1";
    echo "* WARNING: Use default value SCADOPTRUN_N_THREADS_OPENMP_START=$SCADOPTRUN_N_THREADS_OPENMP_START"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_TIMESTEPS" == "" ] ; then
    SCADOPTRUN_N_TIMESTEPS="1";
    echo "* WARNING: Use default value SCADOPTRUN_N_TESTS_MPI=$SCADOPTRUN_N_TIMESTEPS"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_SNAPSHOTS" == "" ] ; then
    SCADOPTRUN_N_SNAPSHOTS="1";
    echo "* WARNING: Use default value SCADOPTRUN_N_SNAPSHOTS=$SCADOPTRUN_N_SNAPSHOTS"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_START_TIME" == "" ] ; then
    SCADOPTRUN_START_TIME="0";
    echo "* WARNING: Use default value SCADOPTRUN_START_TIME=$SCADOPTRUN_START_TIME"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_END_TIME" == "" ] ; then
    SCADOPTRUN_END_TIME="1";
    echo "* WARNING: Use default value SCADOPTRUN_END_TIME=$SCADOPTRUN_END_TIME"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_NAME_CONFIGFILE" == "" ] ; then
    SCADOPTRUN_NAME_CONFIGFILE="";
    echo "* WARNING: Use default value SCADOPTRUN_NAME_CONFIGFILE=$SCADOPTRUN_NAME_CONFIGFILE"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_NAME_KINDFILE" == "" ] ; then
    SCADOPTRUN_NAME_KINDFILE="";
    echo "* WARNING: Use default value SCADOPTRUN_NAME_KINDFILE=$SCADOPTRUN_NAME_KINDFILE"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_WRITE_KINDFILE" == "" ] ; then
    SCADOPTRUN_WRITE_KINDFILE="NO";
    echo "* WARNING: Use default value SCADOPTRUN_WRITE_KINDFILE=$SCADOPTRUN_WRITE_KINDFILE"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_ENABLE_ADOLC" == "" ] ; then
    SCADOPTRUN_ENABLE_ADOLC="NO";
    echo "* WARNING: Use default value SCADOPTRUN_ENABLE_ADOLC=$SCADOPTRUN_ENABLE_ADOLC"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_ASYNCHRON_MODE" == "" ] ; then
    SCADOPTRUN_ASYNCHRON_MODE="YES";
    echo "* WARNING: Use default value SCADOPTRUN_ASYNCHRON_MODE=$SCADOPTRUN_ASYNCHRON_MODE"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_COMP_GRADIENTS" == "" ] ; then
    SCADOPTRUN_COMP_GRADIENTS="NO";
    echo "* WARNING: Use default value SCADOPTRUN_COMP_GRADIENTS=$SCADOPTRUN_COMP_GRADIENTS"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_OPTIM_ITERATIONS_MAX" == "" ] ; then
    SCADOPTRUN_N_OPTIM_ITERATIONS_MAX="50";
    echo "* WARNING: Use default value SCADOPTRUN_N_OPTIM_ITERATIONS_MAX=$SCADOPTRUN_N_OPTIM_ITERATIONS_MAX"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_TF_TIME_DEPENDENT" == "" ] ; then
    SCADOPTRUN_TF_TIME_DEPENDENT="NO";
    echo "* WARNING: Use default value SCADOPTRUN_TF_TIME_DEPENDENT=$SCADOPTRUN_TF_TIME_DEPENDENT"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_RUN_MODE" == "" ] ; then
    SCADOPTRUN_RUN_MODE="NORMAL";
    echo "* WARNING: Use default value SCADOPTRUN_RUN_MODE=$SCADOPTRUN_RUN_MODE"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_OUTPUT_LEVEL" == "" ] ; then
    SCADOPTRUN_OUTPUT_LEVEL="3";
    echo "* WARNING: Use default value SCADOPTRUN_OUTPUT_LEVEL=$SCADOPTRUN_OUTPUT_LEVEL"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_MACHINE_N_NODES" == "" ] ; then
    SCADOPTRUN_MACHINE_N_NODES="1";
    echo "* WARNING: Use default value SCADOPTRUN_MACHINE_N_NODES=$SCADOPTRUN_MACHINE_N_NODES"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_MACHINE_NAME" == "" ] ; then
    SCADOPTRUN_MACHINE_NAME="LOCAL";
    echo "* WARNING: Use default value SCADOPTRUN_MACHINE_NAME=$SCADOPTRUN_MACHINE_NAME"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_MACHINE_N_CORES_PER_NODE" == "" ] ; then
    SCADOPTRUN_MACHINE_N_CORES_PER_NODE="1";
    echo "* WARNING: Use default value SCADOPTRUN_MACHINE_N_CORES_PER_NODE=$SCADOPTRUN_MACHINE_N_CORES_PER_NODE"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_NLAYERS_AT_BORDER" == "" ] ; then
    SCADOPTRUN_NLAYERS_AT_BORDER="1";
    echo "* WARNING: Use default value SCADOPTRUN_NLAYERS_AT_BORDER=$SCADOPTRUN_NLAYERS_AT_BORDER"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_RESULTS_DIR" == "" ] ; then
    SCADOPTRUN_RESULTS_DIR="$PWD";
    echo "* WARNING: Use default value SCADOPTRUN_RESULTS_DIR=$SCADOPTRUN_RESULTS_DIR"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_CHECK_RESULTS" == "" ] ; then
    SCADOPTRUN_CHECK_RESULTS="NO";
    echo "* WARNING: Use default value SCADOPTRUN_CHECK_RESULTS=$SCADOPTRUN_CHECK_RESULTS"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_PARTITIONS" == "" ] ; then
    SCADOPTRUN_N_PARTITIONS="$SCADOPTRUN_N_PROCESSES_MPI_START";
    echo "* WARNING: Use default value SCADOPTRUN_N_PARTITIONS=$SCADOPTRUN_N_PARTITIONS"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_N_NODES" == "" ] ; then
    SCADOPTRUN_N_NODES="8";
    for idxDim in `seq 2 $spaceDim`; do
        tmpNnodes="${SCADOPTRUN_N_NODES}x8";
        SCADOPTRUN_N_NODES=$tmpNnodes;
    done
    echo "* WARNING: Use default value SCADOPTRUN_N_NODES=$SCADOPTRUN_N_NODES"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_COORD_NODE_FIRST" == "" ] ; then
    SCADOPTRUN_COORD_NODE_FIRST="0";
    for idxDim in `seq 2 $spaceDim`; do
        tmpCf="${SCADOPTRUN_COORD_NODE_FIRST}x0";
        SCADOPTRUN_COORD_NODE_FIRST=$tmpCf;
    done
    echo "* WARNING: Use default value SCADOPTRUN_COORD_NODE_FIRST=$SCADOPTRUN_COORD_NODE_FIRST"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_COORD_NODE_LAST" == "" ] ; then
    SCADOPTRUN_COORD_NODE_LAST="1";
    for idxDim in `seq 2 $spaceDim`; do
        tmpCl="${SCADOPTRUN_COORD_NODE_LAST}x1";
        SCADOPTRUN_COORD_NODE_LAST=$tmpCl;
    done
    echo "* WARNING: Use default value SCADOPTRUN_COORD_NODE_LAST=$SCADOPTRUN_COORD_NODE_LAST"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_TYPE_DOMAIN_DECOMPOSITION" == "" ] ; then
    SCADOPTRUN_TYPE_DOMAIN_DECOMPOSITION="RCB";
    echo "* WARNING: Use default value SCADOPTRUN_TYPE_DOMAIN_DECOMPOSITION=$SCADOPTRUN_TYPE_DOMAIN_DECOMPOSITION"
fi

#------------------------------------------------------------------------------#
# Only relevant if SCADOPTRUN_TYPE_DOMAIN_DECOMPOSITION="RCB"; is set.
if [ "$SCADOPTRUN_PARTITION_GRID" == "" ] ; then
    SCADOPTRUN_PARTITION_GRID="1";
    for idxDim in `seq 2 $spaceDim`; do
        tmpPg="${SCADOPTRUN_PARTITION_GRID}x1";
        SCADOPTRUN_PARTITION_GRID=$tmpPg;
    done
    echo "* WARNING: Use default value SCADOPTRUN_PARTITION_GRID=$SCADOPTRUN_PARTITION_GRID"
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_MACHINE_MEMORY_PER_CORE" == "" ] ; then
    if [ "$SCADOPTRUN_RUN_MODE" == "LSF" -o "$SCADOPTRUN_RUN_MODE" == "SLURM" ] ; then
        echo "* ERROR: SCADOPTRUN_MACHINE_MEMORY_PER_CORE was not set."
        exit 22;
    else
        SCADOPTRUN_MACHINE_MEMORY_PER_CORE="900";
        echo "* WARNING: Use default value SCADOPTRUN_MACHINE_MEMORY_PER_CORE=$SCADOPTRUN_MACHINE_MEMORY_PER_CORE";
    fi
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_MACHINE_NAME_PROJECT" == "" ] ; then
    if [ "$SCADOPTRUN_RUN_MODE" == "LSF" -o "$SCADOPTRUN_RUN_MODE" == "SLURM" ] ; then
        echo "* ERROR: SCADOPTRUN_MACHINE_NAME_PROJECT was not set."
        exit 22;
    else
        SCADOPTRUN_MACHINE_NAME_PROJECT="";
        echo "* WARNING: Use default value SCADOPTRUN_MACHINE_NAME_PROJECT=$SCADOPTRUN_MACHINE_NAME_PROJECT"
    fi
fi

#------------------------------------------------------------------------------#
if [ "$SCADOPTRUN_MACHINE_NAME_RESERVATION" == "" ] ; then
    if [ "$SCADOPTRUN_RUN_MODE" == "LSF" -o "$SCADOPTRUN_RUN_MODE" == "SLURM" ] ; then
        echo "* ERROR: SCADOPTRUN_MACHINE_NAME_RESERVATION was not set."
        exit 22;
    else
        SCADOPTRUN_MACHINE_NAME_RESERVATION="";
        echo "* WARNING: Use default value SCADOPTRUN_MACHINE_NAME_RESERVATION=$SCADOPTRUN_MACHINE_NAME_RESERVATION"
    fi
fi

#------------------------------------------------------------------------------#
# From here on, all variables should be declared.
set -u

#------------------------------------------------------------------------------#
# All SCADOPTRUN* environment variables were exported as strings.
# Thus, those SCADOPTRUN* variables which should be used as integer variables
# have to be declared explicitly as local integers variables.
declare -i local nTestsMpi=$SCADOPTRUN_N_TESTS_MPI;
declare -i local nTestsOpenMp=$SCADOPTRUN_N_TESTS_OPENMP;
declare    local testForWeakScaling=$SCADOPTRUN_TYPE_SCALINGTEST;
declare -i local nProcessesMpiStart=$SCADOPTRUN_N_PROCESSES_MPI_START;
declare -i local nThreadsOpenMpStart=$SCADOPTRUN_N_THREADS_OPENMP_START;
declare -i local nTimesteps=$SCADOPTRUN_N_TIMESTEPS;
declare -i local nSnapshots=$SCADOPTRUN_N_SNAPSHOTS;
declare -i local starttime=$SCADOPTRUN_START_TIME;
declare -i local endtime=$SCADOPTRUN_END_TIME; #TODO: Must be set by hand.
declare    local nameConfigfile=$SCADOPTRUN_NAME_CONFIGFILE;
declare    local nameKindfile=$SCADOPTRUN_NAME_KINDFILE;
declare    local writeKindfile=$SCADOPTRUN_WRITE_KINDFILE;
declare    local enableAdolc=$SCADOPTRUN_ENABLE_ADOLC;
declare    local asynchronMode=$SCADOPTRUN_ASYNCHRON_MODE;
declare    local computeGradients=$SCADOPTRUN_COMP_GRADIENTS;
declare    local isTfTimeDependent=$SCADOPTRUN_TF_TIME_DEPENDENT;
declare    local runMode=$SCADOPTRUN_RUN_MODE;
declare    local outputlevel=$SCADOPTRUN_OUTPUT_LEVEL;
declare -i local oneAsInteger=1;
declare -i local nNodesMachineMAX=$SCADOPTRUN_MACHINE_N_NODES;
declare -i local nCoresPerNodeMAX=$SCADOPTRUN_MACHINE_N_CORES_PER_NODE;
declare    local nameMachine=${SCADOPTRUN_MACHINE_NAME};
declare    local softwareVersion="SCADOPTCC_2_0_0";
declare    local nLayersAtBorder="${SCADOPTRUN_NLAYERS_AT_BORDER}";
declare    local resDir=${SCADOPTRUN_RESULTS_DIR};
declare    local checkResults=${SCADOPTRUN_CHECK_RESULTS};
declare    local nPartitions=${SCADOPTRUN_N_PARTITIONS};
declare    local typeDomainDecomposition=${SCADOPTRUN_TYPE_DOMAIN_DECOMPOSITION};
declare    local nameProject=${SCADOPTRUN_MACHINE_NAME_PROJECT};
declare    local nameReservation=${SCADOPTRUN_MACHINE_NAME_RESERVATION};
declare    local memoryPerCore=${SCADOPTRUN_MACHINE_MEMORY_PER_CORE};

#------------------------------------------------------------------------------#
# Compute value of dependent variable.
declare -i local nCoresTotalMAX=${nNodesMachineMAX};
nCoresTotalMAX=$((nCoresTotalMAX*nCoresPerNodeMAX));

#------------------------------------------------------------------------------#
if [ "$runMode" == "DEBUG" ] ; then
    nTestsMpi=1;
    nProcessesMpiStart=1;
    #nThreadsOpenMpStart=1;
fi
#------------------------------------------------------------------------------#
if [ "$nameKindfile" == "" ] ; then
    optionKindfile="";
else
    optionKindfile="--kindfile=$nameKindfile";
fi
#------------------------------------------------------------------------------#
if [ "$nameConfigfile" == "" ] ; then
    optionConfigfile="";
else
    optionConfigfile="--configfile=$nameConfigfile";
fi
#------------------------------------------------------------------------------#
boolEnabledAdolc=0;
if [ "$enableAdolc" == "YES" ] ; then
    boolEnabledAdolc=1;
fi
boolAsynchronMode=0;
if [ "$asynchronMode" == "YES" ] ; then
    boolAsynchronMode=1;
fi
boolCompGradients=0;
if [ "$computeGradients" == "YES" ] ; then
    boolCompGradients=1;
fi
boolIsTfTimeDependent=0;
if [ "$isTfTimeDependent" == "YES" ] ; then
    boolIsTfTimeDependent=1;
fi
boolWriteKindfile=0;
if [ "$writeKindfile" == "YES" ] ; then
    boolWriteKindfile=1;
fi

declare -i local endTestMpi;
declare -i local endTestOpenMp;
declare    local currNameProblem;
declare -i local currIdxProblem;
declare -i local currTest;
declare -i local currIdTest;
declare -i local currIdTestMpi;
declare -i local currIdTestOpenMp;
declare -i local currNprocessesMpi;
declare -i local currNthreadsOpenMp;
declare -i local currEndTest;
declare    local currNameExecutable; # Relative path to executable
declare -i tmpNodesDir;
declare    local errCode=0;
endTestMpi=${nTestsMpi}-${oneAsInteger};
endTestOpenMp=${nTestsOpenMp}-${oneAsInteger};
currEndTest=${nTestsMpi}*${nTestsOpenMp};

timestamp=$(date +%Y%m%d%H%M);
declare local nNodesMachine=1;

currNameProblem=${nameProblem};
currNameExecutable="./${currNameProblem}";
currIdTest=$oneAsInteger;
(
#------------------------------------------------------------------------------#
# Perform a loop over all MPI tests.
currNprocessesMpi=${nProcessesMpiStart};
for idxTestMpi in `seq 0 $endTestMpi`; do
        currIdTestMpi=${idxTestMpi}+${oneAsInteger};

        #----------------------------------------------------------------------#
        # Perform a loop over all OpenMP tests.
        currNthreadsOpenMp=${nThreadsOpenMpStart};
        for idxTestOpenMp in `seq 0 $endTestOpenMp`; do
            currIdTestOpenMp=${idxTestOpenMp}+${oneAsInteger};
            #currNnodes=(${SCADOPTRUN_N_NODES[*]})
            currStarttime=${starttime};
            currEndtime=${endtime};
            currNtimesteps=${nTimesteps};
            currNsnapshots=${nSnapshots};

            #------------------------------------------------------------------#
            divideGrid=${SCADOPTRUN_PARTITION_GRID};
            currNnodes="${SCADOPTRUN_N_NODES}";
            currCoordNodeFirst="${SCADOPTRUN_COORD_NODE_FIRST}";
            currCoordNodeLast="${SCADOPTRUN_COORD_NODE_LAST}";
            namePartfile="${currNameProblem}_${currNprocessesMpi}";
            namePartfile="${namePartfile}_${currNthreadsOpenMp}_${currNnodes}";
            nameDatafile=${namePartfile};

            #------------------------------------------------------------------#
            str_currNprocessesMpi=$(printf "%.5d" $currNprocessesMpi);
            str_currNthreadsOpenMp=$(printf "%.5d" $currNthreadsOpenMp);
            str_nTimesteps=$(printf "%.6d" $currNtimesteps);
            basefile="${resDir}/${currNameProblem}"
            basefile="$basefile-${nameMachine}"
            basefile="$basefile-${softwareVersion}"
            basefile="$basefile-mpi${str_currNprocessesMpi}"
            basefile="$basefile-omp${str_currNthreadsOpenMp}"
            basefile="$basefile-nNodes${currNnodes}"
            basefile="$basefile-decomp${typeDomainDecomposition}"
            basefile="$basefile-nPart${nPartitions}"
            basefile="$basefile-divide${divideGrid}"
            basefile="$basefile-adolc${enableAdolc}"
            basefile="$basefile-async${asynchronMode}"
            basefile="$basefile-grad${computeGradients}"
            basefile="$basefile-tfTdep${isTfTimeDependent}"
            basefile="$basefile-nTs${str_nTimesteps}"
            basefile="$basefile-tEnd${endtime}"
            basefile="$basefile-nLayers${nLayersAtBorder}"
            basefile="$basefile-${timestamp}"
            outputfile="$basefile.out"
            errorfile="$basefile.err"

            #------------------------------------------------------------------#
            if ( expr ${currNprocessesMpi} \< ${nNodesMachineMAX} > /dev/null ) ; then
                nNodesMachine=1;
            else
                nNodesMachine=$((currNprocessesMpi/nNodesMachineMAX));
            fi

            declare -i local nCoresTotal=$((currNprocessesMpi*currNthreadsOpenMp));
            if [ ${nCoresTotal} -gt ${nCoresTotalMAX} ]; then
                echo "* WARNING: nCoresTotal > nCoresTotalMAX. Continue loop."
                continue;
            fi

            #------------------------------------------------------------------#
            # Set default value if nPartitions was not given at the command
            # line: nPartitions=nProcsMpix1x...x1
            # default value: nProcsMpix1x...x1
            declare local isGivenInPartitionGridFormat="no";
            case "$nPartitions" in
               *x*) isGivenInPartitionGridFormat="yes" ;;
            esac

            if [ "$isGivenInPartitionGridFormat" == "no" ] ; then
                 nPartitions="$currNprocessesMpi";
                 for idxDim in `seq 2 $spaceDim`; do
                     tmpPart="${nPartitions}x1"
                     nPartitions=$tmpPart;
                 done
            fi

            echo -e "\n\n  +-------------------------------+";
            echo -e "  | TEST: ${currIdTest} / ${currEndTest}";
            printf "  |     * Name of Problem     = %s\n" ${currNameProblem};
            printf "  |     * #{MPI processes}    = %s\n" ${currNprocessesMpi};
            printf "  |     * #{OpenMP threads}   = %s\n" ${currNthreadsOpenMp};
            printf "  |     * #{Nodes}            = %s\n" ${currNnodes};
            printf "  |     * #{Partitions}       = %s\n" ${nPartitions};
            printf "  |     * Type domain decomp. = %s\n" ${typeDomainDecomposition};
            printf "  |     * divideGrid          = %s\n" ${divideGrid};
            printf "  |     * enableAdolc         = %s\n" ${enableAdolc};
            printf "  |     * asynchronMode       = %s\n" ${asynchronMode};
            printf "  |     - computeGradients    = %s\n" ${computeGradients};
            printf "  |     - isTfTimeDependent   = %s\n" ${isTfTimeDependent};
            printf "  |     - nOptimIterationsMax = %s\n" ${SCADOPTRUN_N_OPTIM_ITERATIONS_MAX};
            printf "  |     * #{Start time}       = %s\n" ${currStarttime};
            printf "  |     * #{End time}         = %s\n" ${currEndtime};
            printf "  |     * #{Time steps}       = %s\n" ${currNtimesteps};
            printf "  |     * #{Layers At Border} = %s\n" ${nLayersAtBorder};
            printf "  |     * #{Snap shots}       = %s\n  |\n" ${currNsnapshots};
            printf "  |     * #{NodesMachineUsed} = %s\n" ${nNodesMachine};
            printf "  |     * #{CoresTotalUsed}   = %s\n  |\n" \
                        $((currNprocessesMpi*currNthreadsOpenMp));

            printf "  |     * Name of Machine     = %s\n" ${nameMachine};
            printf "  |     * #{NodesMachineMAX}  = %s\n" ${nNodesMachineMAX};
            printf "  |     * #{CoresPerNodeMAX}  = %s\n" ${nCoresPerNodeMAX};
            printf "  |     * #{CoresTotalMAX}    = %s\n" ${nCoresTotalMAX};
            echo -e "  +-------------------------------+\n\n";

            #------------------------------------------------------------------#
            echo "--- Run SCADOPT application. ---"
            optsString=" \
                    --dim=$spaceDim \
                    --typeDomainDecomposition=$typeDomainDecomposition \
                    --divideGrid=$divideGrid \
                    --nPartitions=$nPartitions \
                    --nNodes=$currNnodes \
                    --coordNodeFirst=$currCoordNodeFirst \
                    --coordNodeLast=$currCoordNodeLast \
                    --starttime=$currStarttime \
                    --endtime=$currEndtime \
                    --nTimesteps=$currNtimesteps \
                    --nSnapshots=$currNsnapshots \
                    ${optionKindfile} \
                    --writeKindFile=$boolWriteKindfile \
                    ${optionConfigfile} \
                    --enabledAdolc=$boolEnabledAdolc \
                    --asynchronMode=$boolAsynchronMode \
                    --writePartitionfile=$namePartfile \
                    --outputfile=$nameDatafile \
                    --outputlevel=$outputlevel \
                    --nLayersAtBorder=$nLayersAtBorder"
            optsScadoptString=" \
                    --computeGradients=$boolCompGradients \
                    --isTfTimeDependent=$boolIsTfTimeDependent \
                    --nOptimIterationsMax=$SCADOPTRUN_N_OPTIM_ITERATIONS_MAX"
            exeString="$currNameExecutable $optsString $optsScadoptString"

            case $runMode in
            #------------------------------------------------------------------#
            DEBUG )
                (
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                libtool --mode=execute gdb --args $exeString
                )
                ;;
            #------------------------------------------------------------------#
            DEBUG_MPI )
                (
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                libtool --mode=execute mpirun --mca orte_base_helpggregate 0 \
                        -np $currNprocessesMpi \
                        xterm -e gdb --args $exeString
                )
                ;;
            #------------------------------------------------------------------#
            SLURM_WITH_SBATCH_CALL )
                (
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                srun -v -v \
                     --ntasks=$currNprocessesMpi \
                     --cpus-per-task=$OMP_NUM_THREADS \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            NORMAL )
                (
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                mpirun -np $currNprocessesMpi \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            NORMAL_PERF_REPORT )
                (
                 set -x
                 export OMP_NUM_THREADS=$currNthreadsOpenMp;
                 /scratch/zihforschung/allinea/bin/perf-report \
                    mpirun -np $currNprocessesMpi \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            SLURM_PERF_REPORT )
                (
                 set -x
                 export DDT_NO_TIMEOUT=1;
                 export OMP_NUM_THREADS=$currNthreadsOpenMp;
                 # Attention: perf-report:
                 # For 128 MPI processes:
                 # Your current licence (xxx) does not permit as many
                 # processes as you requested.
                 /scratch/zihforschung/allinea/bin/perf-report \
                    -n $currNprocessesMpi --mpi="slurm" \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            LSF )
                (
                # -npernode 64
                set -x
                export OMP_NUM_THREADS=$currNthreadsOpenMp;
                #  -U ${nameReservation}
                bsub \
                  -P ${nameProject} \
                  -n ${nCoresTotal} \
                  -o ${outputfile}.lsf.out \
                  -e ${errorfile}.lsf.err \
                  -J ${currNameExecutable} \
                  -M ${memoryPerCore} \
                  mpirun -np $currNprocessesMpi \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            SLURM )
                (
                ## --reservation=${nameReservation}
                set -x
                srun -v -v \
                    --nodes=${nNodesMachine} \
                    --ntasks=${currNprocessesMpi} \
                    --reservation=${nameReservation} \
                    -o ${outputfile}.slurm.out \
                    -e ${errorfile}.slurm.err \
                    -J ${currNameExecutable} \
                    -A ${nameProject} \
                    --time=01:00:00 \
                    --mem-per-cpu=${memoryPerCore} \
                    --partition=sandy \
                    $exeString > ${outputfile} 2> ${errorfile}
                )
                ;;
            #------------------------------------------------------------------#
            * )
                echo "ERROR: Mode not known." ;;
            esac
            errCode=$?
            echo -e "";
            if [ "${errCode}" != "0" ]; then
                echo -e "RUN TIME ERROR CODE = $errCode. Exit!";
                echo -e "RUN TIME ERROR = YES";
                break 2;
            else
                echo -e "* RUN SUCCESSFULLY?    YES";
            fi
            #------------------------------------------------------------------#
            if [ "$checkResults" == "YES" ]; then
                grep -h "Local Linf" ${outputfile} > tmp.dat
                correctErrCode=$?
                if [ "$correctErrCode" != "0" ]; then
                    echo "CORRECTNESS CHECK: FILE NOT FOUND. Exit!";
                    errCode=1;
                    break 2;
                fi

                grep -h -v "0.0000" tmp.dat
                correctErrCode=$?
                # If lines with error!=0 were found.
                if [ "$correctErrCode" == "0" ]; then
                    echo "* ARE RESULTS CORRECT? NO";
                    errCode=1;
                    break 2;
                else
                    echo "* ARE RESULTS CORRECT? YES";
                    errCode=0;
                fi
            else
                echo "* ARE RESULTS CORRECT? UNKNOWN (DID NOT CHECK)";
            fi

            currNthreadsOpenMp=$currNthreadsOpenMp*2;
            currIdTest=$currIdTest+$oneAsInteger;
        done

        #------------------------------------------------------------------#
        # Test for weak scalability:
        # Problem size increases linearly with number of processes.
        #------------------------------------------------------------------#
        # Test for strong scalability:
        # Problem size remains constant.
        # The number of processes increases linearly.
        if [ "$testForWeakScaling" == "WEAK" ]; then
            echo "TODO: nNodesTotal has to be doubled!";
        else
            currNprocessesMpi=$currNprocessesMpi*2;
            nPartitions="$currNprocessesMpi";
            for idxDim in `seq 2 $spaceDim`; do
                tmpPart="${nPartitions}x1"
                nPartitions=$tmpPart;
            done
        fi
done
exit $errCode;
)
)
